#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging

from . import active_connection as ac
from . import connection
from . import nmclient


def activate(dev=None, connection_id=None):
    """Activate the given device or remote connection profile."""
    conn = connection.ConnectionProfile()
    conn.nmdevice = dev
    conn.con_id = connection_id
    conn.activate()


def deactivate(dev):
    """
    Deactivating the current active connection,
    The profile itself is not removed.

    For software devices, deactivation removes the devices from the kernel.
    """
    act_con = ac.ActiveConnection()
    act_con.nmdevice = dev
    act_con.deactivate()


def delete(dev):
    connections = dev.get_available_connections()
    for con in connections:
        con_profile = connection.ConnectionProfile(con)
        con_profile.delete()


def reapply(dev, connection_profile):
    mainloop = nmclient.mainloop()
    mainloop.push_action(_safe_reapply_async, dev, connection_profile)


def _safe_reapply_async(dev, connection_profile):
    mainloop = nmclient.mainloop()
    cancellable = mainloop.new_cancellable()

    if _wait_for_active_connection_async(dev, connection_profile):
        return

    version_id = 0
    flags = 0
    user_data = mainloop, dev, cancellable
    dev.reapply_async(
        connection_profile,
        version_id,
        flags,
        cancellable,
        _reapply_callback,
        user_data,
    )


def _wait_for_active_connection_async(dev, connection_profile):
    active_conn = connection.get_device_active_connection(dev)
    if active_conn:
        act_conn = ac.ActiveConnection(active_conn)
        if act_conn.is_activating:
            logging.debug(
                'Connection activation in progress: dev=%s, state=%s',
                act_conn.devname,
                act_conn.state,
            )
            conn = connection.ConnectionProfile(connection_profile)
            conn.waitfor_active_connection_async(act_conn)
            return True
    return False


def _reapply_callback(src_object, result, user_data):
    mainloop, nmdev, cancellable = user_data
    mainloop.drop_cancellable(cancellable)

    devname = src_object.get_iface()
    try:
        success = src_object.reapply_finish(result)
    except Exception as e:
        if mainloop.is_action_canceled(e):
            logging.debug('Device reapply aborted on %s: error=%s', devname, e)
        else:
            mainloop.quit(
                'Device reapply failed on {}: error={}'.format(devname, e)
            )
        return

    if success:
        logging.debug('Device reapply succeeded: dev=%s', devname)
        mainloop.execute_next_action()
    else:
        mainloop.quit(
            'Device reapply failed: dev={}, error=unknown'.format(devname)
        )


def modify(dev, connection_profile):
    """
    Modify the given connection profile on the device.
    Implemented by the reapply operation with a fallback to the
    connection profile activation.
    """
    mainloop = nmclient.mainloop()
    mainloop.push_action(_safe_modify_async, dev, connection_profile)


def _safe_modify_async(dev, connection_profile):
    # Special cases are known to exist (bugs) where reapply is not functioning
    # correctly. Use activation instead. (https://bugzilla.redhat.com/1702657)
    if _requires_activation(dev, connection_profile):
        _activate_async(dev)
        return

    mainloop = nmclient.mainloop()
    cancellable = mainloop.new_cancellable()

    if _wait_for_active_connection_async(dev, connection_profile):
        return

    version_id = 0
    flags = 0
    user_data = mainloop, dev, cancellable
    dev.reapply_async(
        connection_profile,
        version_id,
        flags,
        cancellable,
        _modify_callback,
        user_data,
    )


def _modify_callback(src_object, result, user_data):
    mainloop, nmdev, cancellable = user_data
    mainloop.drop_cancellable(cancellable)

    devname = src_object.get_iface()
    try:
        success = src_object.reapply_finish(result)
    except Exception as e:
        if mainloop.is_action_canceled(e):
            logging.debug('Device reapply aborted on %s: error=%s', devname, e)
        else:
            logging.debug(
                'Device reapply failed on %s: error=%s\n'
                'Fallback to device activation',
                devname,
                e,
            )
            _activate_async(src_object)
        return

    if success:
        logging.debug('Device reapply succeeded: dev=%s', devname)
        mainloop.execute_next_action()
    else:
        logging.debug(
            'Device reapply failed, fallback to device activation: dev=%s, '
            'error=unknown',
            devname,
        )
        _activate_async(src_object)


def _requires_activation(dev, connection_profile):
    wired_setting = connection_profile.get_setting_wired()
    configured_mtu = wired_setting.props.mtu if wired_setting else None
    if configured_mtu:
        current_mtu = int(dev.get_mtu())
        if configured_mtu != current_mtu:
            logging.debug(
                'Device reapply does not support mtu changes, '
                'fallback to device activation: dev=%s',
                dev.get_iface(),
            )
            return True

    return False


def _activate_async(dev):
    conn = connection.ConnectionProfile()
    conn.nmdevice = dev
    conn.safe_activate_async()


def delete_device(nmdev):
    mainloop = nmclient.mainloop()
    mainloop.push_action(_safe_delete_device_async, nmdev)


def _safe_delete_device_async(nmdev):
    mainloop = nmclient.mainloop()
    if nmdev.get_state() == nmclient.NM.DeviceState.DEACTIVATING:
        # Nothing to do since the device is already being removed.
        mainloop.execute_next_action()
        return

    user_data = mainloop, nmdev
    nmdev.delete_async(
        mainloop.cancellable, _delete_device_callback, user_data
    )


def _delete_device_callback(src_object, result, user_data):
    mainloop, nmdev = user_data
    try:
        success = src_object.delete_finish(result)
    # pylint: disable=catching-non-exception
    except nmclient.GLib.GError as e:
        # pylint: enable=catching-non-exception
        if e.matches(
            nmclient.Gio.DBusError.quark(),
            nmclient.Gio.DBusError.UNKNOWN_METHOD,
        ):
            logging.debug(
                'Device %s has been already deleted: error=%s',
                nmdev.get_iface(),
                e,
            )
            mainloop.execute_next_action()
        else:
            mainloop.quit(
                'Device deletion failed on {}: error={}'.format(
                    nmdev.get_iface(), e
                )
            )
        return
    except Exception as e:
        if mainloop.is_action_canceled(e):
            logging.debug(
                'Device deletion aborted on %s: error=%s', nmdev.get_iface(), e
            )
        else:
            mainloop.quit(
                'Device deletion failed on {}: error={}'.format(
                    nmdev.get_iface(), e
                )
            )
        return

    devname = src_object.get_iface()
    if success:
        logging.debug('Device deletion succeeded: dev=%s', devname)
        mainloop.execute_next_action()
    else:
        mainloop.quit(
            'Device deletion failed: dev={}, error=unknown'.format(devname)
        )


def get_device_by_name(devname):
    client = nmclient.client()
    return client.get_device_by_iface(devname)


def list_devices():
    client = nmclient.client()
    return client.get_devices()


def get_device_common_info(dev):
    return {
        'name': dev.get_iface(),
        'type_id': dev.get_device_type(),
        'type_name': dev.get_type_description(),
        'state': dev.get_state(),
    }



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from collections import OrderedDict
from copy import deepcopy
import difflib
import json
from operator import itemgetter

import six
import yaml

from libnmstate.schema import Constants


def format_desired_current_state_diff(desired_state, current_state):
    pretty_desired_state = PrettyState(desired_state).yaml
    pretty_current_state = PrettyState(current_state).yaml

    diff = ''.join(
        difflib.unified_diff(
            pretty_desired_state.splitlines(True),
            pretty_current_state.splitlines(True),
            fromfile='desired',
            tofile='current',
            n=3,
        )
    )
    return (
        '\n'
        'desired\n'
        '=======\n'
        '{}\n'
        'current\n'
        '=======\n'
        '{}\n'
        'difference\n'
        '==========\n'
        '{}\n'.format(pretty_desired_state, pretty_current_state, diff)
    )


class PrettyState(object):
    def __init__(self, state):
        yaml.add_representer(OrderedDict, represent_ordereddict)

        if six.PY2:
            yaml.add_representer(six.text_type, represent_unicode)
        self.state = order_state(deepcopy(state))

    @property
    def yaml(self):
        return yaml.dump(
            self.state, default_flow_style=False, explicit_start=True
        )

    @property
    def json(self):
        return json.dumps(self.state, indent=4, separators=(',', ': '))


def represent_ordereddict(dumper, data):
    """
    Represent OrderedDict as regular dictionary

    Source: https://stackoverflow.com/questions/16782112/can-pyyaml-dump-dict-items-in-non-alphabetical-order
    """  # noqa: E501
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


def order_state(state):
    iface_states = state.pop(Constants.INTERFACES, None)

    state = order_iface_state(state)

    if iface_states is not None:
        state[Constants.INTERFACES] = [
            order_iface_state(iface_state)
            for iface_state in sorted(iface_states, key=itemgetter('name'))
        ]

    return state


def represent_unicode(_, data):
    """
    Represent unicode as regular string

    Source:
        https://stackoverflow.com/questions/1950306/pyyaml-dumping-without-tags

    """

    return yaml.ScalarNode(
        tag=u'tag:yaml.org,2002:str', value=data.encode('utf-8')
    )


def order_iface_state(iface_state):
    ordered_state = OrderedDict()

    for setting in ('name', 'type', 'state'):
        try:
            ordered_state[setting] = iface_state.pop(setting)
        except KeyError:
            pass

    for key, value in order_dict(iface_state).items():
        ordered_state[key] = value

    return ordered_state


def order_dict(dict_):
    ordered_dict = OrderedDict()
    for key, value in sorted(dict_.items()):
        if isinstance(value, dict):
            value = order_dict(value)
        ordered_dict[key] = value

    return ordered_dict



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

_IPV6_LINK_LOCAL_NETWORK_PREFIXES = ['fe8', 'fe9', 'fea', 'feb']
_IPV6_LINK_LOCAL_NETWORK_PREFIX_LENGTH = 10

KERNEL_MAIN_ROUTE_TABLE_ID = 254


def is_ipv6_link_local_addr(ip, prefix):
    return (
        ip[: len(_IPV6_LINK_LOCAL_NETWORK_PREFIXES[0])]
        in _IPV6_LINK_LOCAL_NETWORK_PREFIXES
        and prefix >= _IPV6_LINK_LOCAL_NETWORK_PREFIX_LENGTH
    )


def is_ipv6_address(addr):
    return ':' in addr



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
import time
import uuid

from . import nmclient
from .active_connection import ActiveConnection


class ConnectionProfile(object):
    def __init__(self, profile=None):
        self._con_profile = profile
        self._nmclient = nmclient.client()
        self._mainloop = nmclient.mainloop()
        self._nmdevice = None
        self._con_id = None

    def create(self, settings):
        self.profile = nmclient.NM.SimpleConnection.new()
        for setting in settings:
            self.profile.add_setting(setting)

    def import_by_device(self, nmdev=None):
        ac = get_device_active_connection(nmdev or self.nmdevice)
        if ac:
            if nmdev:
                self.nmdevice = nmdev
            self.profile = ac.props.connection

    def import_by_id(self, con_id=None):
        if con_id:
            self.con_id = con_id
        if self.con_id:
            self.profile = self._nmclient.get_connection_by_id(self.con_id)

    def update(self, con_profile):
        self.profile.replace_settings_from_connection(con_profile.profile)

    def add(self, save_to_disk=True):
        user_data = self._mainloop
        self._mainloop.push_action(
            self._nmclient.add_connection_async,
            self.profile,
            save_to_disk,
            self._mainloop.cancellable,
            self._add_connection_callback,
            user_data,
        )

    def delete(self):
        self._mainloop.push_action(self._safe_delete_async)

    def _safe_delete_async(self):
        if not self.profile:
            self.import_by_id()
            if not self.profile:
                self.import_by_device()
        if not self.profile:
            # No callback is expected, so we should call the next one.
            self._mainloop.execute_next_action()
            return

        user_data = None
        self.profile.delete_async(
            self._mainloop.cancellable,
            self._delete_connection_callback,
            user_data,
        )

    def commit(self, save_to_disk=True, nmdev=None):
        user_data = self._mainloop, nmdev
        self._mainloop.push_action(
            self.profile.commit_changes_async,
            save_to_disk,
            self._mainloop.cancellable,
            self._commit_changes_callback,
            user_data,
        )

    def activate(self):
        self._mainloop.push_action(self.safe_activate_async)

    @property
    def profile(self):
        return self._con_profile

    @profile.setter
    def profile(self, con_profile):
        assert self._con_profile is None
        self._con_profile = con_profile

    @property
    def devname(self):
        if self._con_profile:
            return self._con_profile.get_interface_name()
        return None

    @property
    def nmdevice(self):
        return self._nmdevice

    @nmdevice.setter
    def nmdevice(self, dev):
        assert self._nmdevice is None
        self._nmdevice = dev

    @property
    def con_id(self):
        return self._con_id

    @con_id.setter
    def con_id(self, connection_id):
        assert self._con_id is None
        self._con_id = connection_id

    def safe_activate_async(self):
        if self.con_id:
            self.import_by_id()
        elif self.nmdevice:
            self.import_by_device()
        elif not self.profile:
            err_format = 'Missing base properties: profile={}, id={}, dev={}'
            err_msg = err_format.format(
                self.profile, self.con_id, self.devname
            )
            self._mainloop.quit(err_msg)

        cancellable = self._mainloop.new_cancellable()

        active_conn = get_device_active_connection(self.nmdevice)
        if active_conn:
            ac = ActiveConnection(active_conn)
            if ac.is_activating:
                logging.debug(
                    'Connection activation in progress: dev=%s, state=%s',
                    ac.devname,
                    ac.state,
                )
                self.waitfor_active_connection_async(ac)
                return

        specific_object = None
        user_data = cancellable
        self._nmclient.activate_connection_async(
            self.profile,
            self.nmdevice,
            specific_object,
            cancellable,
            self._active_connection_callback,
            user_data,
        )

    def _active_connection_callback(self, src_object, result, user_data):
        cancellable = user_data
        self._mainloop.drop_cancellable(cancellable)

        try:
            nm_act_con = src_object.activate_connection_finish(result)
        except Exception as e:
            act_type, act_object = self._get_activation_metadata()

            if self._mainloop.is_action_canceled(e):
                logging.debug(
                    'Connection activation canceled on %s %s: error=%s',
                    act_type,
                    act_object,
                    e,
                )
            elif self._is_connection_unavailable(e):
                logging.warning(
                    'Connection unavailable on %s %s, retrying',
                    act_type,
                    act_object,
                )
                self._reset_profile()
                time.sleep(0.1)
                self.safe_activate_async()
            else:
                self._mainloop.quit(
                    'Connection activation failed on {} {}: error={}'.format(
                        act_type, act_object, e
                    )
                )
            return

        if nm_act_con is None:
            act_type, act_object = self._get_activation_metadata()
            self._mainloop.quit(
                'Connection activation failed on {} {}: error=unknown'.format(
                    act_type, act_object
                )
            )
        else:
            devname = nm_act_con.props.connection.get_interface_name()
            logging.debug(
                'Connection activation initiated: dev=%s, con-state=%s',
                devname,
                nm_act_con.props.state,
            )

            ac = ActiveConnection(nm_act_con)
            if ac.is_active:
                self._mainloop.execute_next_action()
            elif ac.is_activating:
                self.waitfor_active_connection_async(ac)
            else:
                self._mainloop.quit(
                    'Connection activation failed on {}: reason={}'.format(
                        ac.devname, ac.reason
                    )
                )

    @staticmethod
    def _is_connection_unavailable(err):
        return (
            isinstance(err, nmclient.GLib.GError)
            and err.domain == 'nm-manager-error-quark'
            and err.code == 2
            and 'is not available on the device' in err.message
        )

    def _get_activation_metadata(self):
        if self._nmdevice:
            activation_type = 'device'
            activation_object = self._nmdevice.get_iface()
        elif self._con_id:
            activation_type = 'connection_id'
            activation_object = self._con_id
        else:
            activation_type = activation_object = 'unknown'

        return activation_type, activation_object

    def waitfor_active_connection_async(self, ac):
        ac.handlers.add(
            ac.nm_active_connection.connect(
                'state-changed', self._waitfor_active_connection_callback, ac
            )
        )

    def _waitfor_active_connection_callback(
        self, _nm_act_con, _state, _reason, ac
    ):
        ac.refresh_state()
        if ac.is_active:
            logging.debug(
                'Connection activation succeeded: dev=%s, con-state=%s',
                ac.devname,
                ac.state,
            )
            for handler_id in ac.handlers:
                ac.nm_active_connection.handler_disconnect(handler_id)
            self._mainloop.execute_next_action()
        elif not ac.is_activating:
            for handler_id in ac.handlers:
                ac.nm_active_connection.handler_disconnect(handler_id)
            self._mainloop.quit(
                'Connection activation failed on {}: reason={}'.format(
                    ac.devname, ac.reason
                )
            )

    @staticmethod
    def _add_connection_callback(src_object, result, user_data):
        mainloop = user_data
        try:
            con = src_object.add_connection_finish(result)
        except Exception as e:
            if mainloop.is_action_canceled(e):
                logging.debug('Connection adding canceled: error=%s', e)
            else:
                mainloop.quit('Connection adding failed: error={}'.format(e))
            return

        if con is None:
            mainloop.quit('Connection adding failed: error=unknown')
        else:
            devname = con.get_interface_name()
            logging.debug('Connection adding succeeded: dev=%s', devname)
            mainloop.execute_next_action()

    def _delete_connection_callback(self, src_object, result, user_data):
        try:
            success = src_object.delete_finish(result)
        except Exception as e:
            if self.nmdevice:
                target = 'dev/' + str(self.nmdevice.get_iface())
            else:
                target = 'con/' + str(self.con_id)

            if self._mainloop.is_action_canceled(e):
                logging.debug(
                    'Connection deletion aborted on %s: error=%s', target, e
                )
            else:
                self._mainloop.quit(
                    'Connection deletion failed on {}: error={}'.format(
                        target, e
                    )
                )
            return

        devname = src_object.get_interface_name()
        if success:
            logging.debug('Connection deletion succeeded: dev=%s', devname)
            self._mainloop.execute_next_action()
        else:
            self._mainloop.quit(
                'Connection deletion failed: '
                'dev={}, error=unknown'.format(devname)
            )

    @staticmethod
    def _commit_changes_callback(src_object, result, user_data):
        mainloop, nmdev = user_data
        try:
            success = src_object.commit_changes_finish(result)
        except Exception as e:
            if mainloop.is_action_canceled(e):
                logging.debug('Connection update aborted: error=%s', e)
            else:
                mainloop.quit(
                    'Connection update failed: error={}, dev={}/{}'.format(
                        e, nmdev.props.interface, nmdev.props.state
                    )
                )
            return

        devname = src_object.get_interface_name()
        if success:
            logging.debug('Connection update succeeded: dev=%s', devname)
            mainloop.execute_next_action()
        else:
            mainloop.quit(
                'Connection update failed: '
                'dev={}, error=unknown'.format(devname)
            )

    def _reset_profile(self):
        self._con_profile = None


class ConnectionSetting(object):
    def __init__(self, con_setting=None):
        self._setting = con_setting

    def create(self, con_name, iface_name, iface_type):
        con_setting = nmclient.NM.SettingConnection.new()
        con_setting.props.id = con_name
        con_setting.props.interface_name = iface_name
        con_setting.props.uuid = str(uuid.uuid4())
        con_setting.props.type = iface_type
        con_setting.props.autoconnect = True
        con_setting.props.autoconnect_slaves = (
            nmclient.NM.SettingConnectionAutoconnectSlaves.YES
        )

        self._setting = con_setting
        self._log_connection_info('ConnectionSetting.create')

    def import_by_profile(self, con_profile):
        base = con_profile.profile.get_setting_connection()
        new = nmclient.NM.SettingConnection.new()
        new.props.id = base.props.id
        new.props.interface_name = base.props.interface_name
        new.props.uuid = base.props.uuid
        new.props.type = base.props.type
        new.props.autoconnect = True
        new.props.autoconnect_slaves = base.props.autoconnect_slaves

        self._setting = new
        self._log_connection_info('ConnectionSetting.import_by_profile')

    def set_master(self, master, slave_type):
        if master is not None:
            self._setting.props.master = master
            self._setting.props.slave_type = slave_type

    @property
    def setting(self):
        return self._setting

    def _log_connection_info(self, source):
        setting = self._setting
        logging.debug(
            'Connection settings for %s:\n'
            + '\n'.join(
                [
                    'id: %s',
                    'iface: %s',
                    'uuid: %s',
                    'type: %s',
                    'autoconnect: %s',
                    'autoconnect_slaves: %s',
                ]
            ),
            source,
            setting.props.id,
            setting.props.interface_name,
            setting.props.uuid,
            setting.props.type,
            setting.props.autoconnect,
            setting.props.autoconnect_slaves,
        )


def get_device_active_connection(nm_device):
    active_conn = None
    if nm_device:
        active_conn = nm_device.get_active_connection()
    return active_conn



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import six

from . import connection
from . import nmclient
from libnmstate.error import NmstateValueError


BOND_TYPE = 'bond'


def create_setting(options):
    bond_setting = nmclient.NM.SettingBond.new()
    for option_name, option_value in six.viewitems(options):
        success = bond_setting.add_option(option_name, option_value)
        if not success:
            raise NmstateValueError(
                'Invalid bond option: \{}\=\'{}\''.format(
                    option_name, option_value
                )
            )

    return bond_setting


def is_bond_type_id(type_id):
    return type_id == nmclient.NM.DeviceType.BOND


def get_bond_info(nm_device):
    return {'slaves': get_slaves(nm_device), 'options': get_options(nm_device)}


def get_options(nm_device):
    con = connection.ConnectionProfile()
    con.import_by_device(nm_device)
    if con.profile:
        bond_settings = con.profile.get_setting_bond()
        return bond_settings.props.options
    return {}


def get_slaves(nm_device):
    return nm_device.get_slaves()



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

from collections import deque
from contextlib import contextmanager
import logging

import six

import gi

try:
    gi.require_version('NM', '1.0')  # NOQA: F402
    from gi.repository import NM  # pylint: disable=no-name-in-module
except ValueError:
    NM = None

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject

from libnmstate import error

GObject

_mainloop = None
_nmclient = None


def client(refresh=False):
    global _nmclient
    # refresh is a workaround to get the current state when GMainLoop is not
    # running
    if _nmclient is None or refresh:
        if NM:
            _nmclient = NM.Client.new(None)
        else:
            logging.error(
                'Missing introspection data for libnm'
                'please make sure to install it prior to running nmstate.\n'
                'Check the documentation for more information.'
            )
            raise error.NmstateDependencyError(
                'Missing introspection data for libnm'
            )
    return _nmclient


def mainloop(refresh=False):
    """
    Create a singleton (global) mainloop object.

    When the refresh flag is set, the existing mainloop object is re-created.
    Use the refresh flag in scenarios where the mainloop is needed but
    the existing object is no longer usable (in an undefined state),
    e.g. the mainloop exited with failure.
    """
    global _mainloop
    if _mainloop is None or refresh:
        _mainloop = _MainLoop()
    return _mainloop


class _MainLoop(object):
    SUCCESS = True
    FAIL = False
    RUN_TIMEOUT_ERROR = 'run timeout'
    RUN_EXECUTION_ERROR = 'run execution'

    def __init__(self):
        self._action_queue = deque()
        self._mainloop = GLib.MainLoop()
        self._cancellables = []
        self.new_cancellable()
        self._error = ''

    def execute_next_action(self):
        action = self.pop_action()
        if action:
            func, args, kwargs = action
            logging.debug('Executing NM action: func=%s', func.__name__)
            func(*args, **kwargs)
        else:
            logging.debug('NM action queue exhausted, quiting mainloop')
            self._mainloop.quit()

    def push_action(self, func, *args, **kwargs):
        action = (func, args, kwargs)
        self._action_queue.append(action)

    def pop_action(self):
        try:
            action = self._action_queue.popleft()
        except IndexError:
            return None

        return action if action[0] else None

    def actions_exists(self):
        return bool(self._action_queue)

    def run(self, timeout):
        if not isinstance(timeout, six.integer_types):
            raise error.NmstateValueError(
                "Invalid timeout value: should be an integer"
            )

        if not self.actions_exists():
            return _MainLoop.SUCCESS

        with self._idle_timeout(timeout):
            self._register_first_action()
            self._mainloop.run()

        if self._error == _MainLoop.RUN_TIMEOUT_ERROR:
            return _MainLoop.FAIL

        if len(self._action_queue):
            self._error = _MainLoop.RUN_EXECUTION_ERROR
            return _MainLoop.FAIL

        return _MainLoop.SUCCESS

    @property
    def cancellable(self):
        return self._cancellables[0]

    def new_cancellable(self):
        c = Gio.Cancellable.new()
        self._cancellables.append(c)
        return c

    def drop_cancellable(self, c):
        idx = self._cancellables.index(c)
        if idx == 0:
            raise error.NmstateInternalError('Cannot drop main cancellable')
        del self._cancellables[idx]

    def _cancel_cancellables(self):
        for c in self._cancellables:
            c.cancel()

    def quit(self, reason):
        logging.error('NM main-loop aborted: %s', reason)
        # In case it was the last action, add a sentinel to fail run.
        self.push_action(None)
        self._mainloop.quit()
        self._cancel_cancellables()

    def is_action_canceled(self, err):
        return (
            isinstance(err, GLib.GError)
            and err.domain == 'g-io-error-quark'
            and err.code == Gio.IOErrorEnum.CANCELLED
        )

    @property
    def error(self):
        return self._error

    @contextmanager
    def _idle_timeout(self, timeout):
        timeout_result = []
        data = (self._mainloop, timeout_result)
        timeout_id = GLib.timeout_add(
            int(timeout * 1000), _MainLoop._timeout_cb, data
        )
        yield
        if timeout_result:
            self._cancel_cancellables()
            self._error = _MainLoop.RUN_TIMEOUT_ERROR
        else:
            GLib.source_remove(timeout_id)

    @staticmethod
    def _timeout_cb(data):
        mainloop, result = data
        result.append(1)
        logging.warning('NM main-loop timed out.')
        mainloop.quit()
        return _MainLoop.FAIL

    def _register_first_action(self):
        GLib.timeout_add(1, self._execute_action_once, None)

    def _execute_action_once(self, _):
        self.execute_next_action()
        return False



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

try:
    from collections.abc import Mapping
    from collections.abc import Sequence
except ImportError:
    from collections import Mapping
    from collections import Sequence

from collections import defaultdict
import copy
from functools import total_ordering
from operator import itemgetter

import six

from libnmstate import iplib
from libnmstate import metadata
from libnmstate.error import NmstateValueError
from libnmstate.error import NmstateVerificationError
from libnmstate.iplib import is_ipv6_address
from libnmstate.prettystate import format_desired_current_state_diff
from libnmstate.schema import DNS
from libnmstate.schema import Ethernet
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import Route


NON_UP_STATES = (InterfaceState.DOWN, InterfaceState.ABSENT)


@total_ordering
class RouteEntry(object):
    def __init__(self, route):
        self.table_id = route.get(Route.TABLE_ID)
        self.state = route.get(Route.STATE)
        self.metric = route.get(Route.METRIC)
        self.destination = route.get(Route.DESTINATION)
        self.next_hop_address = route.get(Route.NEXT_HOP_ADDRESS)
        self.next_hop_interface = route.get(Route.NEXT_HOP_INTERFACE)
        self.complement_defaults()

    def complement_defaults(self):
        if self.state != Route.STATE_ABSENT:
            if self.table_id is None:
                self.table_id = Route.USE_DEFAULT_ROUTE_TABLE
            if self.metric is None:
                self.metric = Route.USE_DEFAULT_METRIC
            if self.next_hop_address is None:
                self.next_hop_address = ''

    def __hash__(self):
        return hash(self.__keys())

    def __keys(self):
        return (
            self.table_id,
            self.metric,
            self.destination,
            self.next_hop_address,
            self.next_hop_interface,
        )

    def to_dict(self):
        return {
            key.replace('_', '-'): value
            for key, value in six.viewitems(vars(self))
            if value is not None
        }

    def __eq__(self, other):
        return self is other or self.__keys() == other.__keys()

    def __lt__(self, other):
        return (
            self.table_id or Route.USE_DEFAULT_ROUTE_TABLE,
            self.next_hop_interface or '',
            self.destination or '',
        ) < (
            other.table_id or Route.USE_DEFAULT_ROUTE_TABLE,
            other.next_hop_interface or '',
            other.destination or '',
        )

    def __repr__(self):
        return str(self.to_dict())

    @property
    def absent(self):
        return self.state == Route.STATE_ABSENT

    def match(self, other):
        """
        Match self against other. Treat self None attributes as wildcards,
        matching against any value in others.
        Return True for a match, False otherwise.
        """
        for self_value, other_value in zip(self.__keys(), other.__keys()):
            if self_value is not None and self_value != other_value:
                return False
        return True


def create_state(state, interfaces_to_filter=None):
    """
    Create a state object, given an initial state.
    interface_filter: Limit the interfaces included in the state to the ones
    mentioned in the list. None implied no filtering.
    """
    new_state = {}
    if interfaces_to_filter is not None:
        origin = State(state)
        iface_names = set(origin.interfaces) & interfaces_to_filter
        filtered_ifaces_state = [
            origin.interfaces[ifname] for ifname in iface_names
        ]
        new_state[Interface.KEY] = filtered_ifaces_state

    return State(new_state)


class State(object):
    def __init__(self, state):
        self._state = copy.deepcopy(state)

        self._ifaces_state = self._index_interfaces_state_by_name()
        self._complement_interface_empty_ip_subtrees()

        self._config_iface_routes = self._index_routes_by_iface()

    def __eq__(self, other):
        return self.state == other.state

    def __hash__(self):
        return hash(self.state)

    def __str__(self):
        return str(self.state)

    def __repr__(self):
        return self.__str__()

    @property
    def state(self):
        self._state[Interface.KEY] = sorted(
            list(six.viewvalues(self._ifaces_state)),
            key=itemgetter(Interface.NAME),
        )
        return self._state

    @property
    def interfaces(self):
        """ Indexed interfaces state """
        return self._ifaces_state

    @property
    def config_iface_routes(self):
        """
        Indexed config routes by next hop interface name. Read only.
        """
        return self._config_iface_routes

    @property
    def config_dns(self):
        dns_conf = self._state.get(DNS.KEY, {}).get(DNS.CONFIG, {})
        return {
            DNS.SERVER: dns_conf.get(DNS.SERVER, []),
            DNS.SEARCH: dns_conf.get(DNS.SEARCH, []),
        }

    def _complement_interface_empty_ip_subtrees(self):
        """ Complement the interfaces states with empty IPv4/IPv6 subtrees. """
        for iface_state in six.viewvalues(self.interfaces):
            for family in (Interface.IPV4, Interface.IPV6):
                if family not in iface_state:
                    iface_state[family] = {}

    def sanitize_ethernet(self, other_state):
        """
        Given the other_state, update the ethernet interfaces state base on
        the other_state ethernet interfaces data.
        Usually the other_state represents the current state.
        If auto-negotiation, speed and duplex settings are not provided,
        but exist in the current state, they need to be set to None
        to not override them with the values from the current settings
        since the current settings are read from the device state and not
        from the actual configuration.  This makes it possible to distinguish
        whether a user specified these values in the later configuration step.
        """
        for ifname, iface_state in six.viewitems(self.interfaces):
            iface_current_state = other_state.interfaces.get(ifname, {})
            if iface_current_state.get(Interface.TYPE) == Ethernet.TYPE:
                ethernet = iface_state.setdefault(Ethernet.CONFIG_SUBTREE, {})
                ethernet.setdefault(Ethernet.AUTO_NEGOTIATION, None)
                ethernet.setdefault(Ethernet.SPEED, None)
                ethernet.setdefault(Ethernet.DUPLEX, None)

    def sanitize_dynamic_ip(self):
        """
        If dynamic IP is enabled and IP address is missing, set an empty
        address list. This assures that the desired state is not complemented
        by the current state address values.
        If dynamic IP is disabled, all dynamic IP options should be removed.
        """
        for iface_state in six.viewvalues(self.interfaces):
            for family in ('ipv4', 'ipv6'):
                ip = iface_state[family]
                if ip.get(InterfaceIP.ENABLED) and (
                    ip.get(InterfaceIP.DHCP) or ip.get(InterfaceIPv6.AUTOCONF)
                ):
                    ip[InterfaceIP.ADDRESS] = []
                else:
                    for dhcp_option in (
                        InterfaceIP.AUTO_ROUTES,
                        InterfaceIP.AUTO_GATEWAY,
                        InterfaceIP.AUTO_DNS,
                    ):
                        ip.pop(dhcp_option, None)

    def verify_interfaces(self, other_state):
        """Verify that the (self) state is a subset of the other_state. """
        self._remove_absent_interfaces()
        self._remove_down_virt_interfaces()

        self._assert_interfaces_included_in(other_state)

        metadata.remove_ifaces_metadata(self)
        other_state.sanitize_dynamic_ip()

        self.merge_interfaces(other_state)

        self.normalize_for_verification()
        other_state.normalize_for_verification()

        self._assert_interfaces_equal(other_state)

    def verify_routes(self, other_state):
        for iface_name, routes in six.viewitems(self.config_iface_routes):
            other_routes = other_state.config_iface_routes.get(iface_name, [])
            if routes != other_routes:
                raise NmstateVerificationError(
                    format_desired_current_state_diff(
                        {Route.KEY: [r.to_dict() for r in routes]},
                        {Route.KEY: [r.to_dict() for r in other_routes]},
                    )
                )

    def verify_dns(self, other_state):
        if self.config_dns != other_state.config_dns:
            raise NmstateVerificationError(
                format_desired_current_state_diff(
                    {DNS.KEY: self.config_dns},
                    {DNS.KEY: other_state.config_dns},
                )
            )

    def normalize_for_verification(self):
        self._clean_sanitize_ethernet()
        self._sort_lag_slaves()
        self._sort_bridge_ports()
        self._canonicalize_ipv6()
        self._remove_iface_ipv6_link_local_addr()
        self._sort_ip_addresses()
        self._capitalize_mac()

    def merge_interfaces(self, other_state):
        """
        Given the self and other states, complete the self state by merging
        the missing parts from the current state.
        The operation is performed on entries that exist in both states,
        entries that appear only on one state are ignored.
        This is a reverse recursive update operation.
        """
        other_state = State(other_state.state)
        for name in six.viewkeys(self.interfaces) & six.viewkeys(
            other_state.interfaces
        ):
            dict_update(other_state.interfaces[name], self.interfaces[name])
            self._ifaces_state[name] = other_state.interfaces[name]

    def merge_routes(self, other_state):
        """
        Given the self and other states, complete the self state by merging
        the routes form the current state.
        The resulting routes are a combination of the current routes and the
        desired routes:
        - Self routes are kept.
        - Other routes are kept if:
          - There are self routes under iface OR
          - Self iface explicitly specified (regardless of routes)
        - Self absent routes overwrite other routes state.
          - Support wildcard for matching the absent routes.
        """
        other_routes = set()
        for ifname, routes in six.viewitems(other_state.config_iface_routes):
            if (
                ifname in self.config_iface_routes
                or self._is_interface_routable(ifname, routes)
            ):
                other_routes |= set(routes)

        self_routes = {
            route
            for routes in six.viewvalues(self.config_iface_routes)
            for route in routes
            if not route.absent
        }

        absent_routes = set()
        for routes in six.viewvalues(self.config_iface_routes):
            for absent_route in (r for r in routes if r.absent):
                absent_routes |= {
                    r for r in other_routes if absent_route.match(r)
                }

        merged_routes = (other_routes | self_routes) - absent_routes
        self._config_routes = [r.to_dict() for r in sorted(merged_routes)]
        # FIXME: Index based on route objects directly.
        self._config_iface_routes = self._index_routes_by_iface()

    def _is_interface_routable(self, ifname, routes):
        """
        And interface is able to support routes if:
        - It exists.
        - It is not DOWN or ABSENT.
        - It is not IPv4/6 disabled (corresponding to the routes).
        """
        ifstate = self.interfaces.get(ifname)
        if not ifstate:
            return False

        iface_up = ifstate.get(Interface.STATE) not in NON_UP_STATES
        if iface_up:
            ipv4_state = ifstate.get(Interface.IPV4, {})
            ipv4_disabled = ipv4_state.get(InterfaceIPv4.ENABLED) is False
            if ipv4_disabled and any(
                not is_ipv6_address(r.destination) for r in routes
            ):
                return False

            ipv6_state = ifstate.get(Interface.IPV6, {})
            ipv6_disabled = ipv6_state.get(InterfaceIPv6.ENABLED) is False
            if ipv6_disabled and any(
                is_ipv6_address(r.destination) for r in routes
            ):
                return False

            return True

        return False

    def merge_dns(self, other_state):
        """
        If DNS is not mentioned in the self state, overwrite it with the other
        DNS entries.
        """
        if not self._state.get(DNS.KEY):
            self._state[DNS.KEY] = {
                DNS.CONFIG: copy.deepcopy(other_state.config_dns)
            }

    def _remove_absent_interfaces(self):
        ifaces = {}
        for ifname, ifstate in six.viewitems(self.interfaces):
            is_absent = ifstate.get(Interface.STATE) == InterfaceState.ABSENT
            if not is_absent:
                ifaces[ifname] = ifstate
        self._ifaces_state = ifaces

    def _remove_down_virt_interfaces(self):
        ifaces = {}
        for ifname, ifstate in six.viewitems(self.interfaces):
            is_virt_down = (
                ifstate.get(Interface.STATE) == InterfaceState.DOWN
                and ifstate.get(Interface.TYPE) in InterfaceType.VIRT_TYPES
            )
            if not is_virt_down:
                ifaces[ifname] = ifstate
        self._ifaces_state = ifaces

    def _index_interfaces_state_by_name(self):
        return {
            iface[Interface.NAME]: iface
            for iface in self._state.get(Interface.KEY, [])
        }

    def _index_routes_by_iface(self):
        iface_routes = defaultdict(list)
        for route in self._config_routes:
            iface_name = route.get(Route.NEXT_HOP_INTERFACE, '')
            iface_routes[iface_name].append(RouteEntry(route))
        for routes in six.viewvalues(iface_routes):
            routes.sort()
        return iface_routes

    def _clean_sanitize_ethernet(self):
        for ifstate in six.viewvalues(self.interfaces):
            ethernet_state = ifstate.get(Ethernet.CONFIG_SUBTREE)
            if ethernet_state:
                for key in (
                    Ethernet.AUTO_NEGOTIATION,
                    Ethernet.SPEED,
                    Ethernet.DUPLEX,
                ):
                    if ethernet_state.get(key, None) is None:
                        ethernet_state.pop(key, None)
                if not ethernet_state:
                    ifstate.pop(Ethernet.CONFIG_SUBTREE, None)

    def _sort_lag_slaves(self):
        for ifstate in six.viewvalues(self.interfaces):
            ifstate.get('link-aggregation', {}).get('slaves', []).sort()

    def _sort_bridge_ports(self):
        for ifstate in six.viewvalues(self.interfaces):
            ifstate.get('bridge', {}).get('port', []).sort(
                key=itemgetter('name')
            )

    def _canonicalize_ipv6(self):
        for ifstate in six.viewvalues(self.interfaces):
            new_state = {
                Interface.IPV6: {
                    InterfaceIPv6.ENABLED: False,
                    InterfaceIPv6.ADDRESS: [],
                }
            }
            dict_update(new_state, ifstate)
            self._ifaces_state[ifstate[Interface.NAME]] = new_state

    def _remove_iface_ipv6_link_local_addr(self):
        for ifstate in six.viewvalues(self.interfaces):
            ifstate['ipv6'][InterfaceIPv6.ADDRESS] = list(
                addr
                for addr in ifstate['ipv6'][InterfaceIPv6.ADDRESS]
                if not iplib.is_ipv6_link_local_addr(
                    addr[InterfaceIPv6.ADDRESS_IP],
                    addr[InterfaceIPv6.ADDRESS_PREFIX_LENGTH],
                )
            )

    def _sort_ip_addresses(self):
        for ifstate in six.viewvalues(self.interfaces):
            for family in ('ipv4', 'ipv6'):
                ifstate[family].get(InterfaceIP.ADDRESS, []).sort(
                    key=itemgetter(InterfaceIP.ADDRESS_IP)
                )

    def _capitalize_mac(self):
        for ifstate in six.viewvalues(self.interfaces):
            mac = ifstate.get(Interface.MAC)
            if mac:
                ifstate[Interface.MAC] = mac.upper()

    def _assert_interfaces_equal(self, current_state):
        for ifname in self.interfaces:
            iface_dstate = self.interfaces[ifname]
            iface_cstate = current_state.interfaces[ifname]
            if not state_match(iface_dstate, iface_cstate):
                raise NmstateVerificationError(
                    format_desired_current_state_diff(
                        self.interfaces[ifname],
                        current_state.interfaces[ifname],
                    )
                )

    def _assert_interfaces_included_in(self, current_state):
        if not (set(self.interfaces) <= set(current_state.interfaces)):
            raise NmstateVerificationError(
                format_desired_current_state_diff(
                    self.interfaces, current_state.interfaces
                )
            )

    @property
    def _config_routes(self):
        return self._state.get(Route.KEY, {}).get(Route.CONFIG, [])

    @_config_routes.setter
    def _config_routes(self, value):
        routes = self._state.get(Route.KEY)
        if not routes:
            routes = self._state[Route.KEY] = {}
        routes[Route.CONFIG] = value


def dict_update(origin_data, to_merge_data):
    """Recursevely performes a dict update (merge)"""

    for key, val in six.viewitems(to_merge_data):
        if isinstance(val, Mapping):
            origin_data[key] = dict_update(origin_data.get(key, {}), val)
        else:
            origin_data[key] = val
    return origin_data


def _validate_routes(
    iface_route_sets,
    iface_enable_states,
    ipv4_enable_states,
    ipv6_enable_states,
):
    """
    Check whether user desire routes next hop to:
        * down/absent interface
        * Non-exit interface
        * IPv4/IPv6 disabled
    """
    for iface_name, route_set in six.viewitems(iface_route_sets):
        if not route_set:
            continue
        iface_enable_state = iface_enable_states.get(iface_name)
        if iface_enable_state is None:
            raise NmstateValueError('Cannot set route to non-exist interface')
        if iface_enable_state != InterfaceState.UP:
            raise NmstateValueError(
                'Cannot set route to {} interface'.format(iface_enable_state)
            )
        # Interface is already check, so the ip enable status should be defined
        ipv4_enabled = ipv4_enable_states[iface_name]
        ipv6_enabled = ipv6_enable_states[iface_name]
        for route_obj in route_set:
            if iplib.is_ipv6_address(route_obj.destination):
                if not ipv6_enabled:
                    raise NmstateValueError(
                        'Cannot set IPv6 route when IPv6 is disabled'
                    )
            elif not ipv4_enabled:
                raise NmstateValueError(
                    'Cannot set IPv4 route when IPv4 is disabled'
                )


def _get_iface_enable_states(desire_state, current_state):
    iface_enable_states = {}
    for iface_name, iface_state in six.viewitems(current_state.interfaces):
        iface_enable_states[iface_name] = iface_state[Interface.STATE]
    for iface_name, iface_state in six.viewitems(desire_state.interfaces):
        if Interface.STATE in iface_state:
            # If desire_state does not have Interface.STATE, it will use
            # current_state settings.
            iface_enable_states[iface_name] = iface_state[Interface.STATE]
    return iface_enable_states


def _get_ip_enable_states(family, desire_state, current_state):
    ip_enable_states = {}
    for iface_name, iface_state in six.viewitems(current_state.interfaces):
        ip_enable_states[iface_name] = iface_state.get(family, {}).get(
            InterfaceIP.ENABLED, False
        )
    for iface_name, iface_state in six.viewitems(desire_state.interfaces):
        ip_enable_state = iface_state.get(family, {}).get(InterfaceIP.ENABLED)
        if ip_enable_state is not None:
            # If desire_state does not have Interface.IPV4/IPV6, it will use
            # current_state settings.
            ip_enable_states[iface_name] = ip_enable_state

    return ip_enable_states


def _route_is_valid(
    route_obj, iface_enable_states, ipv4_enable_states, ipv6_enable_states
):
    """
    Return False when route is next hop to any of these interfaces:
        * Interface not in InterfaceState.UP state.
        * Interface does not exists.
        * Interface has IPv4/IPv6 disabled.
    """
    iface_name = route_obj.next_hop_interface
    iface_enable_state = iface_enable_states.get(iface_name)
    if iface_enable_state != InterfaceState.UP:
        return False
    if iplib.is_ipv6_address(route_obj.destination):
        if not ipv6_enable_states.get(iface_name):
            return False
    else:
        if not ipv4_enable_states.get(iface_name):
            return False
    return True


def _apply_absent_routes(absent_route_sets, iface_route_sets):
    """
    Remove routes based on absent routes and treat missing property as wildcard
    match.
    """
    for absent_route in absent_route_sets:
        absent_iface_name = absent_route.next_hop_interface
        for iface_name, route_set in six.viewitems(iface_route_sets):
            if absent_iface_name and absent_iface_name != iface_name:
                continue
            new_routes = set()
            for route in route_set:
                if not absent_route.match(route):
                    new_routes.add(route)
            iface_route_sets[iface_name] = new_routes


def state_match(desire, current):
    """
    Return True when all values defined in desire equal to value in current,
    else False.
    """
    if isinstance(desire, Mapping):
        return isinstance(current, Mapping) and all(
            state_match(val, current.get(key))
            for key, val in six.viewitems(desire)
        )
    elif isinstance(desire, Sequence) and not isinstance(
        desire, six.string_types
    ):
        return (
            isinstance(current, Sequence)
            and not isinstance(current, six.string_types)
            and len(current) == len(desire)
            and all(state_match(d, c) for d, c in zip(desire, current))
        )
    else:
        return desire == current



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import six

from libnmstate.nm import connection
from libnmstate.nm import nmclient
from libnmstate.schema import LinuxBridge as LB


BRIDGE_TYPE = 'bridge'


def create_setting(options, base_con_profile):
    bridge_setting = _get_current_bridge_setting(base_con_profile)
    if not bridge_setting:
        bridge_setting = nmclient.NM.SettingBridge.new()

    if options:
        _set_bridge_properties(bridge_setting, options)

    return bridge_setting


def _get_current_bridge_setting(base_con_profile):
    bridge_setting = None
    if base_con_profile:
        bridge_setting = base_con_profile.get_setting_bridge()
        if bridge_setting:
            bridge_setting = bridge_setting.duplicate()
    return bridge_setting


def _set_bridge_properties(bridge_setting, options):
    for key, val in six.viewitems(options):
        if key == LB.MAC_AGEING_TIME:
            bridge_setting.props.ageing_time = val
        elif key == LB.GROUP_FORWARD_MASK:
            bridge_setting.props.group_forward_mask = val
        elif key == LB.MULTICAST_SNOOPING:
            bridge_setting.props.multicast_snooping = val
        elif key == LB.STP_SUBTREE:
            _set_bridge_stp_properties(bridge_setting, val)


def _set_bridge_stp_properties(bridge_setting, bridge_stp):
    bridge_setting.props.stp = bridge_stp[LB.STP_ENABLED]
    if bridge_stp[LB.STP_ENABLED] is True:
        for stp_key, stp_val in six.viewitems(bridge_stp):
            if stp_key == LB.STP_PRIORITY:
                bridge_setting.props.priority = stp_val
            elif stp_key == LB.STP_FORWARD_DELAY:
                bridge_setting.props.forward_delay = stp_val
            elif stp_key == LB.STP_HELLO_TIME:
                bridge_setting.props.hello_time = stp_val
            elif stp_key == LB.STP_MAX_AGE:
                bridge_setting.props.max_age = stp_val


def create_port_setting(options, base_con_profile):
    port_setting = None
    if base_con_profile:
        port_setting = base_con_profile.get_setting_bridge_port()
        if port_setting:
            port_setting = port_setting.duplicate()

    if not port_setting:
        port_setting = nmclient.NM.SettingBridgePort.new()

    for key, val in six.viewitems(options):
        if key == LB.PORT_STP_PRIORITY:
            port_setting.props.priority = val
        elif key == LB.PORT_STP_HAIRPIN_MODE:
            port_setting.props.hairpin_mode = val
        elif key == LB.PORT_STP_PATH_COST:
            port_setting.props.path_cost = val

    return port_setting


def get_info(nmdev):
    """
    Provides the current active values for a device
    """
    info = {}
    if nmdev.get_device_type() != nmclient.NM.DeviceType.BRIDGE:
        return info
    bridge_setting = _get_bridge_setting(nmdev)
    if not bridge_setting:
        return info

    port_profiles = _get_slave_profiles(nmdev)
    info[LB.CONFIG_SUBTREE] = {
        LB.PORT_SUBTREE: _get_bridge_ports_info(port_profiles),
        LB.OPTIONS_SUBTREE: {
            LB.MAC_AGEING_TIME: bridge_setting.props.ageing_time,
            LB.GROUP_FORWARD_MASK: bridge_setting.props.group_forward_mask,
            LB.MULTICAST_SNOOPING: bridge_setting.props.multicast_snooping,
            LB.STP_SUBTREE: {
                LB.STP_ENABLED: bridge_setting.props.stp,
                LB.STP_PRIORITY: bridge_setting.props.priority,
                LB.STP_FORWARD_DELAY: bridge_setting.props.forward_delay,
                LB.STP_HELLO_TIME: bridge_setting.props.hello_time,
                LB.STP_MAX_AGE: bridge_setting.props.max_age,
            },
        },
    }
    return info


def get_slaves(nm_device):
    return nm_device.get_slaves()


def _get_bridge_setting(nmdev):
    bridge_setting = None
    bridge_con_profile = connection.ConnectionProfile()
    bridge_con_profile.import_by_device(nmdev)
    if bridge_con_profile.profile:
        bridge_setting = bridge_con_profile.profile.get_setting_bridge()
    return bridge_setting


def _get_bridge_ports_info(port_profiles):
    ports_info = []
    for p in port_profiles:
        port_info = _get_bridge_port_info(p)
        if port_info:
            ports_info.append(port_info)
    return ports_info


def _get_bridge_port_info(port_profile):
    """Report port information."""

    port_setting = port_profile.get_setting_bridge_port()
    return {
        LB.PORT_NAME: port_profile.get_interface_name(),
        LB.PORT_STP_PRIORITY: port_setting.props.priority,
        LB.PORT_STP_HAIRPIN_MODE: port_setting.props.hairpin_mode,
        LB.PORT_STP_PATH_COST: port_setting.props.path_cost,
    }


def _get_slave_profiles(master_device):
    slave_profiles = []
    for dev in master_device.get_slaves():
        active_con = connection.get_device_active_connection(dev)
        if active_con:
            slave_profiles.append(active_con.props.connection)
    return slave_profiles



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

import pkgutil
import yaml


def load(schema_name):
    return yaml.load(
        pkgutil.get_data('libnmstate', 'schemas/' + schema_name + '.yaml'),
        Loader=yaml.SafeLoader,
    )


ifaces_schema = load('operational-state')


class Interface(object):
    KEY = 'interfaces'

    NAME = 'name'
    TYPE = 'type'
    STATE = 'state'

    IPV4 = 'ipv4'
    IPV6 = 'ipv6'

    MAC = 'mac-address'
    MTU = 'mtu'


class Route(object):
    KEY = 'routes'

    RUNNING = 'running'
    CONFIG = 'config'
    STATE = 'state'
    STATE_ABSENT = 'absent'
    TABLE_ID = 'table-id'
    DESTINATION = 'destination'
    NEXT_HOP_INTERFACE = 'next-hop-interface'
    NEXT_HOP_ADDRESS = 'next-hop-address'
    METRIC = 'metric'
    USE_DEFAULT_METRIC = -1
    USE_DEFAULT_ROUTE_TABLE = 0


class DNS(object):
    KEY = 'dns-resolver'
    RUNNING = 'running'
    CONFIG = 'config'
    SERVER = 'server'
    SEARCH = 'search'


class Constants(object):
    INTERFACES = Interface.KEY
    ROUTES = Route.KEY
    DNS = DNS.KEY


class InterfaceState(object):
    KEY = Interface.STATE

    DOWN = 'down'
    UP = 'up'
    ABSENT = 'absent'


class InterfaceType(object):
    KEY = Interface.TYPE

    BOND = 'bond'
    ETHERNET = 'ethernet'
    LINUX_BRIDGE = 'linux-bridge'
    OVS_BRIDGE = 'ovs-bridge'
    OVS_INTERFACE = 'ovs-interface'
    OVS_PORT = 'ovs-port'
    UNKNOWN = 'unknown'
    VLAN = 'vlan'

    VIRT_TYPES = (
        BOND,
        LINUX_BRIDGE,
        OVS_BRIDGE,
        OVS_PORT,
        OVS_INTERFACE,
        VLAN,
    )


class InterfaceIP(object):
    ENABLED = 'enabled'
    ADDRESS = 'address'
    ADDRESS_IP = 'ip'
    ADDRESS_PREFIX_LENGTH = 'prefix-length'
    DHCP = 'dhcp'
    AUTO_DNS = 'auto-dns'
    AUTO_GATEWAY = 'auto-gateway'
    AUTO_ROUTES = 'auto-routes'


class InterfaceIPv4(InterfaceIP):
    pass


class InterfaceIPv6(InterfaceIP):
    AUTOCONF = 'autoconf'


class Bond(object):
    KEY = InterfaceType.BOND
    CONFIG_SUBTREE = 'link-aggregation'

    MODE = 'mode'
    SLAVES = 'slaves'
    OPTIONS_SUBTREE = 'options'


class BondMode(object):
    ROUND_ROBIN = 'balance-rr'
    ACTIVE_BACKUP = 'active-backup'
    XOR = 'balance-xor'
    BROADCAST = 'broadcast'
    LACP = '802.3ad'
    TLB = 'balance-tlb'
    ALB = 'balance-alb'


class LinuxBridge(object):
    TYPE = 'linux-bridge'
    CONFIG_SUBTREE = 'bridge'

    OPTIONS_SUBTREE = 'options'
    MAC_AGEING_TIME = 'mac-ageing-time'
    GROUP_FORWARD_MASK = 'group-forward-mask'
    MULTICAST_SNOOPING = 'multicast-snooping'

    STP_SUBTREE = 'stp'
    STP_ENABLED = 'enabled'
    STP_PRIORITY = 'priority'
    STP_FORWARD_DELAY = 'forward-delay'
    STP_HELLO_TIME = 'hello-time'
    STP_MAX_AGE = 'max-age'

    PORT_SUBTREE = 'port'
    PORT_NAME = 'name'
    PORT_STP_PRIORITY = 'stp-priority'
    PORT_STP_HAIRPIN_MODE = 'stp-hairpin-mode'
    PORT_STP_PATH_COST = 'stp-path-cost'


class Ethernet(object):
    TYPE = InterfaceType.ETHERNET
    CONFIG_SUBTREE = 'ethernet'

    AUTO_NEGOTIATION = 'auto-negotiation'
    SPEED = 'speed'
    DUPLEX = 'duplex'

    FULL_DUPLEX = 'full'
    HALF_DUPLEX = 'half'


class OVSBridge(object):
    TYPE = 'ovs-bridge'
    CONFIG_SUBTREE = 'bridge'

    OPTIONS_SUBTREE = 'options'
    FAIL_MODE = 'fail-mode'
    MCAST_SNOOPING_ENABLED = 'mcast-snooping-enable'
    RSTP = 'rstp'
    STP = 'stp'

    PORT_SUBTREE = 'port'
    PORT_NAME = 'name'
    PORT_TYPE = 'type'
    PORT_VLAN_MODE = 'vlan-mode'
    PORT_ACCESS_TAG = 'access-tag'


class OVSBridgePortType(object):
    SYSTEM = 'system'
    INTERNAL = 'internal'



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

from . import error
from . import schema

from .netapplier import apply
from .netapplier import commit
from .netapplier import rollback
from .netinfo import show

from .prettystate import PrettyState


__all__ = [
    'show',
    'apply',
    'commit',
    'rollback',
    'error',
    'schema',
    'PrettyState',
]



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import six

from libnmstate.error import NmstateValueError
from libnmstate.schema import OVSBridge as OB
from libnmstate.schema import OVSBridgePortType as OBPortType

from . import connection
from . import device
from . import nmclient


BRIDGE_TYPE = 'ovs-bridge'
INTERNAL_INTERFACE_TYPE = 'ovs-interface'
PORT_TYPE = 'ovs-port'
PORT_PROFILE_PREFIX = 'ovs-port-'
CAPABILITY = 'openvswitch'


_BRIDGE_OPTION_NAMES = ['fail-mode', 'mcast-snooping-enable', 'rstp', 'stp']


_PORT_OPTION_NAMES = [
    'tag',
    'vlan-mode',
    'bond-mode',
    'lacp',
    'bond-updelay',
    'bond-downdelay',
]


def has_ovs_capability():
    try:
        nmclient.NM.DeviceType.OVS_BRIDGE
        return True
    except AttributeError:
        return False


def create_bridge_setting(options):
    bridge_setting = nmclient.NM.SettingOvsBridge.new()
    for option_name, option_value in six.viewitems(options):
        if option_name == 'fail-mode':
            if option_value:
                bridge_setting.props.fail_mode = option_value
        elif option_name == 'mcast-snooping-enable':
            bridge_setting.props.mcast_snooping_enable = option_value
        elif option_name == 'rstp':
            bridge_setting.props.rstp_enable = option_value
        elif option_name == 'stp':
            bridge_setting.props.stp_enable = option_value
        else:
            raise NmstateValueError(
                'Invalid OVS bridge option: \{}\=\'{}\''.format(
                    option_name, option_value
                )
            )

    return bridge_setting


def create_port_setting(options):
    port_setting = nmclient.NM.SettingOvsPort.new()
    for option_name, option_value in six.viewitems(options):
        if option_name == 'tag':
            port_setting.props.tag = option_value
        elif option_name == 'vlan-mode':
            port_setting.props.vlan_mode = option_value
        elif option_name == 'bond-mode':
            port_setting.props.bond_mode = option_value
        elif option_name == 'lacp':
            port_setting.props.lacp = option_value
        elif option_name == 'bond-updelay':
            port_setting.props.bond_updelay = option_value
        elif option_name == 'bond-downdelay':
            port_setting.props.bond_downdelay = option_value
        else:
            raise NmstateValueError(
                'Invalid OVS port option: \{}\=\'{}\''.format(
                    option_name, option_value
                )
            )

    return port_setting


def create_interface_setting():
    interface_setting = nmclient.NM.SettingOvsInterface.new()
    interface_setting.props.type = 'internal'
    return interface_setting


def translate_bridge_options(iface_state):
    br_opts = {}
    bridge_state = iface_state.get('bridge', {}).get('options', {})
    for key in six.viewkeys(bridge_state) & set(_BRIDGE_OPTION_NAMES):
        br_opts[key] = bridge_state[key]

    return br_opts


def translate_port_options(port_state):
    port_opts = {}
    for key in six.viewkeys(port_state) & set(_PORT_OPTION_NAMES):
        port_opts[key] = port_state[key]

    return port_opts


def is_ovs_bridge_type_id(type_id):
    return type_id == nmclient.NM.DeviceType.OVS_BRIDGE


def is_ovs_port_type_id(type_id):
    return type_id == nmclient.NM.DeviceType.OVS_PORT


def is_ovs_interface_type_id(type_id):
    return type_id == nmclient.NM.DeviceType.OVS_INTERFACE


def get_bridge_info(bridge_device, devices_info):
    return {OB.CONFIG_SUBTREE: get_ovs_info(bridge_device, devices_info)}


def get_ovs_info(bridge_device, devices_info):
    port_profiles = _get_slave_profiles(bridge_device, devices_info)

    return {
        'port': _get_bridge_ports_info(port_profiles, devices_info),
        'options': _get_bridge_options(bridge_device),
    }


def _get_bridge_ports_info(port_profiles, devices_info):
    ports_info = []
    for p in port_profiles:
        port_info = _get_bridge_port_info(p, devices_info)
        if port_info:
            ports_info.append(port_info)
    return ports_info


def _get_bridge_port_info(port_profile, devices_info):
    """
    Report port information.
    Note: The current implementation supports only system OVS ports and
    access vlan-mode (trunks are not supported).
    """
    port_info = {}

    port_setting = port_profile.get_setting(nmclient.NM.SettingOvsPort)
    vlan_mode = port_setting.props.vlan_mode

    port_name = port_profile.get_interface_name()
    port_device = device.get_device_by_name(port_name)
    port_slave_profiles = _get_slave_profiles(port_device, devices_info)
    port_slave_names = [c.get_interface_name() for c in port_slave_profiles]

    if port_slave_names:
        iface_slave_name = port_slave_names[0]
        iface_device = device.get_device_by_name(iface_slave_name)

        if is_ovs_interface_type_id(iface_device.props.device_type):
            port_type = OBPortType.INTERNAL
        else:
            port_type = OBPortType.SYSTEM

        port_info['name'] = iface_slave_name
        port_info['type'] = port_type
        if vlan_mode:
            port_info['vlan-mode'] = vlan_mode
            port_info['access-tag'] = port_setting.props.tag

    return port_info


def _get_bridge_options(bridge_device):
    bridge_options = {}
    con = connection.ConnectionProfile()
    con.import_by_device(bridge_device)
    if con.profile:
        bridge_setting = con.profile.get_setting(nmclient.NM.SettingOvsBridge)
        bridge_options['stp'] = bridge_setting.props.stp_enable
        bridge_options['rstp'] = bridge_setting.props.rstp_enable
        bridge_options['fail-mode'] = bridge_setting.props.fail_mode or ''
        bridge_options[
            'mcast-snooping-enable'
        ] = bridge_setting.props.mcast_snooping_enable

    return bridge_options


def _get_slave_profiles(master_device, devices_info):
    slave_profiles = []
    for dev, _ in devices_info:
        active_con = connection.get_device_active_connection(dev)
        if active_con:
            master = active_con.props.master
            if master and (master.get_iface() == master_device.get_iface()):
                slave_profiles.append(active_con.props.connection)
    return slave_profiles



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy
import logging
import six

import jsonschema as js

from . import nm
from . import schema
from .schema import Constants
from libnmstate.schema import DNS
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv6
from libnmstate.error import NmstateDependencyError
from libnmstate.error import NmstateNotImplementedError
from libnmstate.error import NmstateValueError
from libnmstate.iplib import is_ipv6_address


class NmstateRouteWithNoInterfaceError(NmstateValueError):
    pass


class NmstateRouteWithNoUpInterfaceError(NmstateValueError):
    pass


class NmstateRouteWithNoIPInterfaceError(NmstateValueError):
    pass


def validate(data, validation_schema=schema.ifaces_schema):
    data = copy.deepcopy(data)
    for ifstate in data.get(schema.Interface.KEY, ()):
        if not ifstate.get(schema.Interface.TYPE):
            ifstate[schema.Interface.TYPE] = schema.InterfaceType.UNKNOWN
    js.validate(data, validation_schema)


def validate_capabilities(state, capabilities):
    validate_interface_capabilities(state[Constants.INTERFACES], capabilities)


def validate_interface_capabilities(ifaces_state, capabilities):
    ifaces_types = [iface_state.get('type') for iface_state in ifaces_state]
    has_ovs_capability = nm.ovs.CAPABILITY in capabilities
    for iface_type in ifaces_types:
        is_ovs_type = iface_type in (
            nm.ovs.BRIDGE_TYPE,
            nm.ovs.PORT_TYPE,
            nm.ovs.INTERNAL_INTERFACE_TYPE,
        )
        if is_ovs_type and not has_ovs_capability:
            raise NmstateDependencyError(
                "Open vSwitch NetworkManager support not installed "
                "and started"
            )


def validate_interfaces_state(desired_state, current_state):
    validate_link_aggregation_state(desired_state, current_state)


def validate_link_aggregation_state(desired_state, current_state):
    available_ifaces = {
        ifname
        for ifname, ifstate in six.viewitems(desired_state.interfaces)
        if ifstate.get('state') != 'absent'
    }
    available_ifaces |= set(current_state.interfaces)

    specified_slaves = set()
    for iface_state in six.viewvalues(desired_state.interfaces):
        if iface_state.get('state') != 'absent':
            link_aggregation = iface_state.get('link-aggregation')
            if link_aggregation:
                slaves = set(link_aggregation.get('slaves', []))
                if not (slaves <= available_ifaces):
                    raise NmstateValueError(
                        "Link aggregation has missing slave: {}".format(
                            iface_state
                        )
                    )
                if slaves & specified_slaves:
                    raise NmstateValueError(
                        "Link aggregation has reused slave: {}".format(
                            iface_state
                        )
                    )
                specified_slaves |= slaves


def validate_dhcp(state):
    for iface_state in state[Constants.INTERFACES]:
        for family in ('ipv4', 'ipv6'):
            ip = iface_state.get(family, {})
            if (
                ip.get(InterfaceIP.ENABLED)
                and ip.get(InterfaceIP.ADDRESS)
                and (
                    ip.get(InterfaceIP.DHCP) or ip.get(InterfaceIPv6.AUTOCONF)
                )
            ):
                logging.warning(
                    '%s addresses are ignored when ' 'dynamic IP is enabled',
                    family,
                )


def validate_dns(state):
    """
    Only support at most 2 name servers now:
    https://nmstate.atlassian.net/browse/NMSTATE-220
    """
    dns_servers = (
        state.get(DNS.KEY, {}).get(DNS.CONFIG, {}).get(DNS.SERVER, [])
    )
    if len(dns_servers) > 2:
        raise NmstateNotImplementedError(
            'Nmstate only support at most 2 DNS name servers'
        )


def validate_routes(desired_state, current_state):
    """
    A route has several requirements it must comply with:
    - Next-hop interface must be provided
    - The next-hop interface must:
        - Exist and be up (no down/absent)
        - Have the relevant IPv4/6 stack enabled.
    """
    for iface_name, routes in six.viewitems(desired_state.config_iface_routes):
        if not routes:
            continue

        desired_iface_state = desired_state.interfaces.get(iface_name)
        current_iface_state = current_state.interfaces.get(iface_name)
        if desired_iface_state or current_iface_state:
            _assert_iface_is_up(desired_iface_state, current_iface_state)
            if any(is_ipv6_address(route.destination) for route in routes):
                _assert_iface_ipv6_enabled(
                    desired_iface_state, current_iface_state
                )
            if any(not is_ipv6_address(route.destination) for route in routes):
                _assert_iface_ipv4_enabled(
                    desired_iface_state, current_iface_state
                )
        else:
            raise NmstateRouteWithNoInterfaceError(str(routes))


def _assert_iface_is_up(desired_iface_state, current_iface_state):
    """
    Validates that the interface has an UP state.
    Prioritize the desired state over the current state.
    """
    if desired_iface_state:
        state = desired_iface_state.get(schema.Interface.STATE)
        if state is not None:
            if state == schema.InterfaceState.UP:
                return
            raise NmstateRouteWithNoUpInterfaceError(desired_iface_state)
    if current_iface_state:
        state = current_iface_state.get(schema.Interface.STATE)
        if state != schema.InterfaceState.UP:
            raise NmstateRouteWithNoUpInterfaceError(current_iface_state)


def _assert_iface_ipv4_enabled(desired_iface_state, current_iface_state):
    _assert_iface_ip_enabled(
        desired_iface_state, current_iface_state, schema.Interface.IPV4
    )


def _assert_iface_ipv6_enabled(desired_iface_state, current_iface_state):
    _assert_iface_ip_enabled(
        desired_iface_state, current_iface_state, schema.Interface.IPV6
    )


def _assert_iface_ip_enabled(desired_iface_state, current_iface_state, ipkey):
    """
    Validates that the interface has IPv4/6 (ipkey) enabled.
    Prioritize the desired state over the current state.
    """
    if desired_iface_state:
        ip_state = desired_iface_state.get(ipkey)
        if ip_state is not None:
            ip_enabled = ip_state.get(InterfaceIP.ENABLED)
            if ip_enabled is True:
                return
            elif ip_enabled is False:
                raise NmstateRouteWithNoIPInterfaceError(desired_iface_state)
    if current_iface_state:
        ip_state = current_iface_state.get(ipkey)
        if not ip_state.get(InterfaceIP.ENABLED):
            raise NmstateRouteWithNoIPInterfaceError(current_iface_state)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from operator import itemgetter

from libnmstate import nm
from libnmstate import validator
from libnmstate.nm import dns as nm_dns
from libnmstate.schema import Constants
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import Route


def show(include_status_data=False):
    """
    Reports configuration and status data on the system.
    Configuration data is the set of writable data which can change the system
    state.
    Status data is the additional data which is not configuration data,
    including read-only and statistics information.
    When include_status_data is set, both are reported, otherwise only the
    configuration data is reported.
    """
    client = nm.nmclient.client(refresh=True)

    report = {Constants.INTERFACES: interfaces()}
    if include_status_data:
        report['capabilities'] = capabilities()

    report[Constants.ROUTES] = {
        Route.RUNNING: (
            nm.ipv4.get_route_running() + nm.ipv6.get_route_running()
        ),
        Route.CONFIG: (
            nm.ipv4.get_route_config() + nm.ipv6.get_route_config()
        ),
    }

    report[Constants.DNS] = {
        DNS.RUNNING: nm_dns.get_running(),
        DNS.CONFIG: nm_dns.get_config(
            nm.ipv4.acs_and_ip_profiles(client),
            nm.ipv6.acs_and_ip_profiles(client),
        ),
    }

    validator.validate(report)
    return report


def capabilities():
    caps = set()

    if nm.ovs.has_ovs_capability():
        caps.add(nm.ovs.CAPABILITY)

    return list(caps)


def interfaces():
    info = []

    nm.nmclient.client(refresh=True)

    devices_info = [
        (dev, nm.device.get_device_common_info(dev))
        for dev in nm.device.list_devices()
    ]

    for dev, devinfo in devices_info:
        type_id = devinfo['type_id']

        iface_info = nm.translator.Nm2Api.get_common_device_info(devinfo)

        act_con = nm.connection.get_device_active_connection(dev)
        iface_info['ipv4'] = nm.ipv4.get_info(act_con)
        iface_info['ipv6'] = nm.ipv6.get_info(act_con)
        iface_info.update(nm.wired.get_info(dev))
        iface_info.update(nm.user.get_info(dev))
        iface_info.update(nm.vlan.get_info(dev))
        iface_info.update(nm.bridge.get_info(dev))

        if nm.bond.is_bond_type_id(type_id):
            bondinfo = nm.bond.get_bond_info(dev)
            iface_info.update(_ifaceinfo_bond(bondinfo))
        elif nm.ovs.has_ovs_capability():
            if nm.ovs.is_ovs_bridge_type_id(type_id):
                iface_info['bridge'] = nm.ovs.get_ovs_info(dev, devices_info)
                iface_info = _remove_ovs_bridge_unsupported_entries(iface_info)
            elif nm.ovs.is_ovs_port_type_id(type_id):
                continue

        info.append(iface_info)

    info.sort(key=itemgetter('name'))

    return info


def _ifaceinfo_bond(devinfo):
    # TODO: What about unmanaged devices?
    bondinfo = nm.translator.Nm2Api.get_bond_info(devinfo)
    if 'link-aggregation' in bondinfo:
        return bondinfo
    return {}


def _remove_ovs_bridge_unsupported_entries(iface_info):
    """
    OVS bridges are not supporting several common interface key entries.
    These entries are removed explicitly.
    """
    iface_info.pop(Interface.IPV4, None)
    iface_info.pop(Interface.IPV6, None)
    iface_info.pop(Interface.MTU, None)

    return iface_info



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from operator import itemgetter
import socket

import six

from libnmstate import iplib
from libnmstate.error import NmstateInternalError
from libnmstate.error import NmstateNotImplementedError
from libnmstate.nm import nmclient
from libnmstate.nm import active_connection as nm_ac
from libnmstate.schema import Interface
from libnmstate.schema import Route

NM_ROUTE_TABLE_ATTRIBUTE = 'table'
IPV4_DEFAULT_GATEWAY_DESTINATION = '0.0.0.0/0'
IPV6_DEFAULT_GATEWAY_DESTINATION = '::/0'
ROUTE_METADATA = '_routes'


def get_running(acs_and_ip_cfgs):
    """
    Query running routes
    The acs_and_ip_cfgs should be generate to generate a tuple:
        NM.NM.ActiveConnection, NM.IPConfig
    """
    routes = []
    for (active_connection, ip_cfg) in acs_and_ip_cfgs:
        if not ip_cfg.props.routes:
            continue
        iface_name = nm_ac.ActiveConnection(active_connection).devname
        if not iface_name:
            raise NmstateInternalError(
                'Got connection {} has not interface name'.format(
                    active_connection.get_id()
                )
            )
        for nm_route in ip_cfg.props.routes:
            table_id = _get_per_route_table_id(
                nm_route, iplib.KERNEL_MAIN_ROUTE_TABLE_ID
            )
            route_entry = _nm_route_to_route(nm_route, table_id, iface_name)
            if route_entry:
                routes.append(route_entry)
    routes.sort(
        key=itemgetter(
            Route.TABLE_ID, Route.NEXT_HOP_INTERFACE, Route.DESTINATION
        )
    )
    return routes


def get_config(acs_and_ip_profiles):
    """
    Query running routes
    The acs_and_ip_profiles should be generate to generate a tuple:
        NM.NM.ActiveConnection, NM.SettingIPConfig
    """
    routes = []
    for (active_connection, ip_profile) in acs_and_ip_profiles:
        nm_routes = ip_profile.props.routes
        gateway = ip_profile.props.gateway
        if not nm_routes and not gateway:
            continue
        iface_name = nm_ac.ActiveConnection(active_connection).devname
        if not iface_name:
            raise NmstateInternalError(
                'Got connection {} has not interface name'.format(
                    active_connection.get_id()
                )
            )
        default_table_id = ip_profile.props.route_table
        if gateway:
            routes.append(
                _get_default_route_config(
                    gateway,
                    ip_profile.props.route_metric,
                    default_table_id,
                    iface_name,
                )
            )
        # NM supports multiple route table in single profile:
        #   https://bugzilla.redhat.com/show_bug.cgi?id=1436531
        # The `ipv4.route-table` and `ipv6.route-table` will be the default
        # table id for static routes and auto routes. But each static route can
        # still specify route table id.
        for nm_route in nm_routes:
            table_id = _get_per_route_table_id(nm_route, default_table_id)
            route_entry = _nm_route_to_route(nm_route, table_id, iface_name)
            if route_entry:
                routes.append(route_entry)
    routes.sort(
        key=itemgetter(
            Route.TABLE_ID, Route.NEXT_HOP_INTERFACE, Route.DESTINATION
        )
    )
    return routes


def _get_per_route_table_id(nm_route, default_table_id):
    table = nm_route.get_attribute(NM_ROUTE_TABLE_ATTRIBUTE)
    return int(table.get_uint32()) if table else default_table_id


def _nm_route_to_route(nm_route, table_id, iface_name):
    dst = '{ip}/{prefix}'.format(
        ip=nm_route.get_dest(), prefix=nm_route.get_prefix()
    )
    next_hop = nm_route.get_next_hop() or ''
    metric = int(nm_route.get_metric())

    return {
        Route.TABLE_ID: table_id,
        Route.DESTINATION: dst,
        Route.NEXT_HOP_INTERFACE: iface_name,
        Route.NEXT_HOP_ADDRESS: next_hop,
        Route.METRIC: metric,
    }


def _get_default_route_config(gateway, metric, default_table_id, iface_name):
    if iplib.is_ipv6_address(gateway):
        destination = IPV6_DEFAULT_GATEWAY_DESTINATION
    else:
        destination = IPV4_DEFAULT_GATEWAY_DESTINATION
    return {
        Route.TABLE_ID: default_table_id,
        Route.DESTINATION: destination,
        Route.NEXT_HOP_INTERFACE: iface_name,
        Route.NEXT_HOP_ADDRESS: gateway,
        Route.METRIC: metric,
    }


def add_routes(setting_ip, routes):
    for route in routes:
        if route[Route.DESTINATION] in (
            IPV4_DEFAULT_GATEWAY_DESTINATION,
            IPV6_DEFAULT_GATEWAY_DESTINATION,
        ):
            if setting_ip.get_gateway():
                raise NmstateNotImplementedError(
                    'Only a single default gateway is supported due to a '
                    'limitation of NetworkManager: '
                    'https://bugzilla.redhat.com/1707396'
                )
            _add_route_gateway(setting_ip, route)
        else:
            _add_specfic_route(setting_ip, route)


def _add_specfic_route(setting_ip, route):
    destination, prefix_len = route[Route.DESTINATION].split('/')
    prefix_len = int(prefix_len)
    if iplib.is_ipv6_address(destination):
        family = socket.AF_INET6
    else:
        family = socket.AF_INET
    metric = route.get(Route.METRIC, Route.USE_DEFAULT_METRIC)
    next_hop = route[Route.NEXT_HOP_ADDRESS]
    ip_route = nmclient.NM.IPRoute.new(
        family, destination, prefix_len, next_hop, metric
    )
    table_id = route.get(Route.TABLE_ID, Route.USE_DEFAULT_ROUTE_TABLE)
    ip_route.set_attribute(
        NM_ROUTE_TABLE_ATTRIBUTE, nmclient.GLib.Variant.new_uint32(table_id)
    )
    # Duplicate route entry will be ignored by libnm.
    setting_ip.add_route(ip_route)


def _add_route_gateway(setting_ip, route):
    setting_ip.props.gateway = route[Route.NEXT_HOP_ADDRESS]
    setting_ip.props.route_table = route.get(
        Route.TABLE_ID, Route.USE_DEFAULT_ROUTE_TABLE
    )
    setting_ip.props.route_metric = route.get(
        Route.METRIC, Route.USE_DEFAULT_METRIC
    )


def get_static_gateway_iface(family, iface_routes):
    """
    Return one interface with gateway for given IP family.
    Return None if not found.
    """
    destination = (
        IPV6_DEFAULT_GATEWAY_DESTINATION
        if family == Interface.IPV6
        else IPV4_DEFAULT_GATEWAY_DESTINATION
    )
    for iface_name, routes in six.viewitems(iface_routes):
        for route in routes:
            if route[Route.DESTINATION] == destination:
                return iface_name
    return None



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy

import six

from . import nmclient


IFACE_TYPE_UNKNOWN = 'unknown'


class ApiIfaceAdminState(object):
    DOWN = 'down'
    UP = 'up'


class Api2Nm(object):
    _iface_types_map = None

    @staticmethod
    def get_iface_type(name):
        return Api2Nm.get_iface_type_map().get(name, IFACE_TYPE_UNKNOWN)

    @staticmethod
    def get_iface_type_map():
        if Api2Nm._iface_types_map is None:
            Api2Nm._iface_types_map = {
                'ethernet': nmclient.NM.SETTING_WIRED_SETTING_NAME,
                'bond': nmclient.NM.SETTING_BOND_SETTING_NAME,
                'dummy': nmclient.NM.SETTING_DUMMY_SETTING_NAME,
                'vlan': nmclient.NM.SETTING_VLAN_SETTING_NAME,
                'linux-bridge': nmclient.NM.SETTING_BRIDGE_SETTING_NAME,
            }
            try:
                ovs_types = {
                    'ovs-bridge': nmclient.NM.SETTING_OVS_BRIDGE_SETTING_NAME,
                    'ovs-port': nmclient.NM.SETTING_OVS_PORT_SETTING_NAME,
                    'ovs-interface': (
                        nmclient.NM.SETTING_OVS_INTERFACE_SETTING_NAME
                    ),
                }
                Api2Nm._iface_types_map.update(ovs_types)
            except AttributeError:
                pass

        return Api2Nm._iface_types_map

    @staticmethod
    def get_bond_options(iface_desired_state):
        iface_type = Api2Nm.get_iface_type(iface_desired_state['type'])
        if iface_type == 'bond':
            # Is the mode a must config parameter?
            bond_conf = iface_desired_state['link-aggregation']
            bond_opts = {'mode': bond_conf['mode']}
            bond_opts.update(bond_conf.get('options', {}))
        else:
            bond_opts = {}

        return bond_opts


class Nm2Api(object):
    _iface_types_map = None

    @staticmethod
    def get_common_device_info(devinfo):
        type_name = devinfo['type_name']
        if type_name != 'ethernet':
            type_name = Nm2Api.get_iface_type(type_name)
        return {
            'name': devinfo['name'],
            'type': type_name,
            'state': Nm2Api.get_iface_admin_state(devinfo['state']),
        }

    @staticmethod
    def get_bond_info(bondinfo):
        bond_options = copy.deepcopy(bondinfo.get('options'))
        if not bond_options:
            return {}
        bond_slaves = bondinfo['slaves']

        bond_mode = bond_options['mode']
        del bond_options['mode']
        return {
            'link-aggregation': {
                'mode': bond_mode,
                'slaves': [slave.props.interface for slave in bond_slaves],
                'options': bond_options,
            }
        }

    @staticmethod
    def get_iface_type(name):
        if Nm2Api._iface_types_map is None:
            Nm2Api._iface_types_map = Nm2Api._swap_dict_keyval(
                Api2Nm.get_iface_type_map()
            )
        return Nm2Api._iface_types_map.get(name, IFACE_TYPE_UNKNOWN)

    @staticmethod
    def get_iface_admin_state(state_name):
        if state_name in (
            nmclient.NM.DeviceState.ACTIVATED,
            nmclient.NM.DeviceState.IP_CONFIG,
        ):
            return ApiIfaceAdminState.UP
        return ApiIfaceAdminState.DOWN

    @staticmethod
    def _swap_dict_keyval(dictionary):
        return {val: key for key, val in six.viewitems(dictionary)}



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import six

from libnmstate import iplib
from libnmstate.appliers import linux_bridge
from libnmstate.error import NmstateValueError
from libnmstate import nm
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceState


BRPORT_OPTIONS = '_brport_options'
MASTER = '_master'
MASTER_TYPE = '_master_type'
ROUTES = '_routes'
DNS_METADATA = '_dns'
DNS_METADATA_PRIORITY = '_priority'


def generate_ifaces_metadata(desired_state, current_state):
    """
    The described desired state for each interface may include references to
    other interfaces. As the provider handles the interface setting in an
    isolated manner, it is sometime necessary to specify a property on an
    interface based on a property from a different interface.
    An exmaple of this, is the bond slaves property, which should mark each
    slave with the setting of it being a slave.

    For such relationships between interfaces or some other potential inputs,
    metadata is generated on interfaces, usable by the provider when
    configuring the interface.
    """
    _generate_link_master_metadata(
        desired_state.interfaces,
        current_state.interfaces,
        master_type='bond',
        get_slaves_func=_get_bond_slaves_from_state,
        set_metadata_func=_set_common_slaves_metadata,
    )
    _generate_link_master_metadata(
        desired_state.interfaces,
        current_state.interfaces,
        master_type='ovs-bridge',
        get_slaves_func=_get_ovs_slaves_from_state,
        set_metadata_func=_set_ovs_bridge_ports_metadata,
    )
    _generate_link_master_metadata(
        desired_state.interfaces,
        current_state.interfaces,
        master_type='linux-bridge',
        get_slaves_func=linux_bridge.get_slaves_from_state,
        set_metadata_func=linux_bridge.set_bridge_ports_metadata,
    )
    _generate_dns_metadata(desired_state, current_state)
    _generate_route_metadata(desired_state, current_state)


def remove_ifaces_metadata(ifaces_state):
    for iface_state in six.viewvalues(ifaces_state.interfaces):
        iface_state.pop(MASTER, None)
        iface_state.pop(MASTER_TYPE, None)
        iface_state.pop(BRPORT_OPTIONS, None)
        iface_state.get(Interface.IPV4, {}).pop(ROUTES, None)
        iface_state.get(Interface.IPV6, {}).pop(ROUTES, None)
        iface_state.get(Interface.IPV4, {}).pop(DNS_METADATA, None)
        iface_state.get(Interface.IPV6, {}).pop(DNS_METADATA, None)


def _get_bond_slaves_from_state(iface_state, default=()):
    return iface_state.get('link-aggregation', {}).get('slaves', default)


def _set_ovs_bridge_ports_metadata(master_state, slave_state):
    _set_common_slaves_metadata(master_state, slave_state)

    ports = master_state.get('bridge', {}).get('port', [])
    port = next(
        six.moves.filter(lambda n: n['name'] == slave_state['name'], ports), {}
    )
    slave_state[BRPORT_OPTIONS] = port


def _set_common_slaves_metadata(master_state, slave_state):
    slave_state[MASTER] = master_state['name']
    slave_state[MASTER_TYPE] = master_state.get('type')


def _get_ovs_slaves_from_state(iface_state, default=()):
    ports = iface_state.get('bridge', {}).get('port')
    if ports is None:
        return default
    return [p['name'] for p in ports]


def _generate_link_master_metadata(
    ifaces_desired_state,
    ifaces_current_state,
    master_type,
    get_slaves_func,
    set_metadata_func,
):
    """
    Given master's slaves, add to the slave interface the master information.

    Possible scenarios for a given desired and current sate:
    - The desired state contains both the masters and their slaves.
    - The desired state contains the masters and partially (or not at all)
      the slaves. Some or all the slaves are in the current state.
    - Master is in the current state and some of the slaves are in the desired
      state.
    """
    desired_masters = [
        (ifname, ifstate)
        for ifname, ifstate in six.viewitems(ifaces_desired_state)
        if ifstate.get('type') == master_type
        and ifstate.get('state') not in ('down', 'absent')
    ]
    for master_name, master_state in desired_masters:
        desired_slaves = get_slaves_func(master_state)
        for slave in desired_slaves:
            if slave in ifaces_desired_state:
                set_metadata_func(master_state, ifaces_desired_state[slave])
            elif slave in ifaces_current_state:
                ifaces_desired_state[slave] = {
                    'name': slave,
                    'state': master_state['state'],
                }
                set_metadata_func(master_state, ifaces_desired_state[slave])

    for master_name, master_state in desired_masters:
        desired_slaves = get_slaves_func(master_state)
        current_master_state = ifaces_current_state.get(master_name)
        if desired_slaves and current_master_state:
            current_slaves = get_slaves_func(current_master_state)
            slaves2remove = set(current_slaves) - set(desired_slaves)
            for slave in slaves2remove:
                if slave not in ifaces_desired_state:
                    ifaces_desired_state[slave] = {
                        'name': slave,
                        'state': 'down',
                    }

    current_masters = (
        (ifname, ifstate)
        for ifname, ifstate in six.viewitems(ifaces_current_state)
        if ifstate.get('type') == master_type
    )
    for master_name, master_state in current_masters:
        current_slaves = get_slaves_func(master_state)
        for slave in current_slaves:
            if slave in ifaces_desired_state:
                iface_state = ifaces_desired_state.get(master_name, {})
                if iface_state.get('state') not in ('down', 'absent'):
                    master_has_no_slaves_specified_in_desired = (
                        get_slaves_func(iface_state, None) is None
                    )
                    slave_has_no_master_specified_in_desired = (
                        ifaces_desired_state[slave].get(MASTER) is None
                    )
                    if (
                        slave_has_no_master_specified_in_desired
                        and master_has_no_slaves_specified_in_desired
                    ):
                        set_metadata_func(
                            master_state, ifaces_desired_state[slave]
                        )


def _generate_route_metadata(desired_state, current_state):
    """
    Save routes metadata under interface IP protocol for future processing by
    the providers.
    Currently route['next-hop-interface'] is mandatory.
    Routes which do not match any current or desired interface are ignored.
    """
    for iface_name, routes in six.viewitems(desired_state.config_iface_routes):
        desired_iface_state = desired_state.interfaces.get(iface_name)
        current_iface_state = current_state.interfaces.get(iface_name)
        if desired_iface_state:
            _attach_route_metadata(desired_iface_state, routes)
        elif current_iface_state:
            desired_iface_state = desired_state.interfaces[iface_name] = {
                Interface.NAME: iface_name,
                Interface.TYPE: current_iface_state[Interface.TYPE],
            }
            _attach_route_metadata(desired_iface_state, routes)


def _attach_route_metadata(iface_state, routes):
    _init_iface_route_metadata(iface_state, Interface.IPV4)
    _init_iface_route_metadata(iface_state, Interface.IPV6)

    for route in routes:
        if iplib.is_ipv6_address(route.destination):
            iface_state[Interface.IPV6][ROUTES].append(route.to_dict())
        else:
            iface_state[Interface.IPV4][ROUTES].append(route.to_dict())


def _init_iface_route_metadata(iface_state, ip_key):
    ip_state = iface_state.get(ip_key)
    if not ip_state:
        ip_state = iface_state[ip_key] = {}
    ip_state[ROUTES] = []


def _generate_dns_metadata(desired_state, current_state):
    """
    Save DNS configuration on chosen interfaces as metadata.
    """
    _clear_current_dns_config(desired_state, current_state)
    servers = desired_state.config_dns.get(DNS.SERVER, [])
    searches = desired_state.config_dns.get(DNS.SEARCH, [])
    if not servers and not searches:
        return
    if _dns_config_not_changed(desired_state, current_state):
        _preserve_current_dns_metadata(desired_state, current_state)
    else:
        ifaces_routes = {
            ifname: [r.to_dict() for r in routes]
            for ifname, routes in six.viewitems(
                desired_state.config_iface_routes
            )
        }
        ipv4_iface, ipv6_iface = nm.dns.find_interfaces_for_name_servers(
            ifaces_routes
        )
        _save_dns_metadata(
            desired_state,
            current_state,
            ipv4_iface,
            ipv6_iface,
            servers,
            searches,
        )


def _save_dns_metadata(
    desired_state, current_state, ipv4_iface, ipv6_iface, servers, searches
):
    index = 0
    searches_saved = False
    for server in servers:
        iface_name = None
        if iplib.is_ipv6_address(server):
            iface_name = ipv6_iface
            family = Interface.IPV6
        else:
            iface_name = ipv4_iface
            family = Interface.IPV4
        if not iface_name:
            raise NmstateValueError(
                'Failed to find suitable interface for saving DNS '
                'name servers: %s' % server
            )

        _include_name_only_iface_state(
            desired_state, current_state, [iface_name]
        )
        iface_state = desired_state.interfaces[iface_name]
        if family not in iface_state:
            iface_state[family] = {}

        if DNS_METADATA not in iface_state[family]:
            iface_state[family][DNS_METADATA] = {
                DNS.SERVER: [server],
                DNS.SEARCH: [] if searches_saved else searches,
                DNS_METADATA_PRIORITY: nm.dns.DNS_PRIORITY_STATIC_BASE + index,
            }
        else:
            iface_state[family][DNS_METADATA][DNS.SERVER].append(server)
        searches_saved = True
        index += 1


def _include_name_only_iface_state(desired_state, current_state, iface_names):
    ifnames = set(iface_names) & set(current_state.interfaces) - set(
        desired_state.interfaces
    )
    for ifname in ifnames:
        desired_state.interfaces[ifname] = {Interface.NAME: ifname}

    for iface_name in iface_names:
        if (
            iface_name not in desired_state.interfaces
            and iface_name in current_state.interfaces
        ):
            desired_state.interfaces[iface_name] = {Interface.NAME: iface_name}


def _clear_current_dns_config(desired_state, current_state):
    client = nm.nmclient.client()
    current_dns_ifaces = nm.dns.get_dns_config_iface_names(
        nm.ipv4.acs_and_ip_profiles(client),
        nm.ipv6.acs_and_ip_profiles(client),
    )
    _include_name_only_iface_state(
        desired_state, current_state, current_dns_ifaces
    )


def _dns_config_not_changed(desired_state, current_state):
    """
    Return True if desired_state DNS config equal to current_state and
    interface holding current DNS config is UP and corresponding IP family
    is enabled.
    """
    if desired_state.config_dns != current_state.config_dns:
        return False
    client = nm.nmclient.client()
    iface_dns_configs = nm.dns.get_indexed_dns_config_by_iface(
        nm.ipv4.acs_and_ip_profiles(client),
        nm.ipv6.acs_and_ip_profiles(client),
    )
    for iface_name, iface_dns_config in six.viewitems(iface_dns_configs):
        if iface_name not in desired_state.interfaces:
            continue
        iface_state = desired_state.interfaces[iface_name]
        if (
            Interface.STATE in iface_state
            and iface_state[Interface.STATE] != InterfaceState.UP
        ):
            return False
        for family in six.viewkeys(iface_dns_config):
            if family in iface_state and not iface_state[family].get(
                InterfaceIP.ENABLED
            ):
                return False
    return True


def _preserve_current_dns_metadata(desired_state, current_state):
    client = nm.nmclient.client()
    iface_dns_configs = nm.dns.get_indexed_dns_config_by_iface(
        nm.ipv4.acs_and_ip_profiles(client),
        nm.ipv6.acs_and_ip_profiles(client),
    )
    for iface_name, iface_dns_config in six.viewitems(iface_dns_configs):
        if iface_name not in desired_state.interfaces:
            continue
        for family, dns_metadata in six.viewitems(iface_dns_config):
            iface_state = desired_state.interfaces[iface_name]
            if family not in iface_state:
                iface_state[family] = {}
            iface_state[family][DNS_METADATA] = dns_metadata



#
# Copyright 2007 Google Inc.
#  Licensed to PSF under a Contributor Agreement.
#
# Copyright 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import ipaddress


class IPv4Network(ipaddress.IPv4Network):
    def subnet_of(self, other):
        """Return True if this network is a subnet of other."""
        return self._is_subnet_of(self, other)

    @staticmethod
    def _is_subnet_of(a, b):
        try:
            # Always false if one is v4 and the other is v6.
            if a._version != b._version:
                raise TypeError(
                    '{} and {} are not of the same version'.format(a, b)
                )
            return (
                b.network_address <= a.network_address
                and b.broadcast_address >= a.broadcast_address
            )
        except AttributeError:
            raise TypeError(
                'Unable to test subnet containment '
                'between {} and {}'.format(a, b)
            )



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import itertools
import six

from libnmstate.error import NmstateValueError
from libnmstate.schema import LinuxBridge as LB

from . import bond
from . import bridge
from . import connection
from . import device
from . import ipv4
from . import ipv6
from . import nmclient
from . import ovs
from . import translator
from . import user
from . import vlan
from . import wired


MASTER_METADATA = '_master'
MASTER_TYPE_METADATA = '_master_type'

BRPORT_OPTIONS_METADATA = '_brport_options'


def create_new_ifaces(con_profiles):
    for connection_profile in con_profiles:
        connection_profile.add(save_to_disk=True)


def prepare_new_ifaces_configuration(ifaces_desired_state):
    return [
        _build_connection_profile(iface_desired_state)
        for iface_desired_state in ifaces_desired_state
    ]


def edit_existing_ifaces(con_profiles):
    for connection_profile in con_profiles:
        devname = connection_profile.devname
        nmdev = device.get_device_by_name(devname)
        cur_con_profile = None
        if nmdev:
            cur_con_profile = connection.ConnectionProfile()
            cur_con_profile.import_by_device(nmdev)
        if cur_con_profile and cur_con_profile.profile:
            connection_profile.commit(nmdev=nmdev)
        else:
            # Missing connection, attempting to create a new one.
            connection_profile.add(save_to_disk=True)


def prepare_edited_ifaces_configuration(ifaces_desired_state):
    con_profiles = []

    for iface_desired_state in ifaces_desired_state:
        nmdev = device.get_device_by_name(iface_desired_state['name'])
        cur_con_profile = None
        if nmdev:
            cur_con_profile = connection.ConnectionProfile()
            cur_con_profile.import_by_device(nmdev)
        new_con_profile = _build_connection_profile(
            iface_desired_state, base_con_profile=cur_con_profile
        )
        if not new_con_profile.devname:
            set_conn = new_con_profile.profile.get_setting_connection()
            set_conn.props.interface_name = iface_desired_state['name']
        if cur_con_profile and cur_con_profile.profile:
            cur_con_profile.update(new_con_profile)
            con_profiles.append(cur_con_profile)
        else:
            # Missing connection, attempting to create a new one.
            con_profiles.append(new_con_profile)

    return con_profiles


def set_ifaces_admin_state(ifaces_desired_state, con_profiles=()):
    """
    Control interface admin state by activating, deactivating and deleting
    devices connection profiles.

    The `absent` state results in deactivating the device and deleting
    the connection profile.
    FIXME: The `down` state is currently handled in the same way.

    For new virtual devices, the `up` state is handled by activating the
    new connection profile. For existing devices, the device is activated,
    leaving it to choose the correct profile.

    In order to activate correctly the interfaces, the order is significant:
    - New interfaces (virtual interfaces, but not OVS ones).
    - Master interfaces.
    - OVS ports.
    - OVS internal.
    - All the rest.
    """
    con_profiles_by_devname = _index_profiles_by_devname(con_profiles)
    new_ifaces = _get_new_ifaces(con_profiles)
    new_ifaces_to_activate = set()
    new_ovs_interface_to_activate = set()
    new_ovs_port_to_activate = set()
    master_ifaces_to_edit = set()
    ifaces_to_edit = set()
    remove_devs_actions = {}

    for iface_desired_state in ifaces_desired_state:
        ifname = iface_desired_state['name']
        nmdev = device.get_device_by_name(ifname)
        if not nmdev:
            if ifname in new_ifaces and iface_desired_state['state'] == 'up':
                if iface_desired_state['type'] == ovs.INTERNAL_INTERFACE_TYPE:
                    new_ovs_interface_to_activate.add(ifname)
                elif iface_desired_state['type'] == ovs.PORT_TYPE:
                    new_ovs_port_to_activate.add(ifname)
                else:
                    new_ifaces_to_activate.add(ifname)
        else:
            if iface_desired_state['state'] == 'up':
                master_iface_types = ovs.BRIDGE_TYPE, bond.BOND_TYPE, LB.TYPE
                if iface_desired_state['type'] in master_iface_types:
                    master_ifaces_to_edit.add(
                        (nmdev, con_profiles_by_devname[ifname].profile)
                    )
                else:
                    ifaces_to_edit.add(
                        (nmdev, con_profiles_by_devname[ifname].profile)
                    )
            elif iface_desired_state['state'] in ('down', 'absent'):
                nmdevs = _get_affected_devices(iface_desired_state)
                for nmdev in nmdevs:
                    remove_devs_actions[nmdev] = [
                        device.deactivate,
                        device.delete,
                    ]
                    if nmdev.get_device_type() in (
                        nmclient.NM.DeviceType.OVS_BRIDGE,
                        nmclient.NM.DeviceType.OVS_PORT,
                        nmclient.NM.DeviceType.OVS_INTERFACE,
                    ):
                        remove_devs_actions[nmdev].append(device.delete_device)
            else:
                raise NmstateValueError(
                    'Invalid state {} for interface {}'.format(
                        iface_desired_state['state'],
                        iface_desired_state['name'],
                    )
                )

    # Do not remove devices that are marked for editing.
    for dev, _ in itertools.chain(master_ifaces_to_edit, ifaces_to_edit):
        remove_devs_actions.pop(dev, None)

    for ifname in new_ifaces_to_activate:
        device.activate(dev=None, connection_id=ifname)

    for dev, con_profile in master_ifaces_to_edit:
        device.modify(dev, con_profile)

    for ifname in new_ovs_port_to_activate:
        device.activate(dev=None, connection_id=ifname)

    for ifname in new_ovs_interface_to_activate:
        device.activate(dev=None, connection_id=ifname)

    for dev, con_profile in ifaces_to_edit:
        device.modify(dev, con_profile)

    for dev, actions in six.viewitems(remove_devs_actions):
        for action in actions:
            action(dev)


def _index_profiles_by_devname(con_profiles):
    return {con_profile.devname: con_profile for con_profile in con_profiles}


def _get_new_ifaces(con_profiles):
    ifaces_without_device = set()
    for con_profile in con_profiles:
        ifname = con_profile.devname
        nmdev = device.get_device_by_name(ifname)
        if not nmdev:
            ifaces_without_device.add(ifname)
    return ifaces_without_device


def _get_affected_devices(iface_state):
    nmdev = device.get_device_by_name(iface_state['name'])
    devs = []
    if nmdev:
        devs += [nmdev]
        iface_type = iface_state['type']
        if iface_type == ovs.BRIDGE_TYPE:
            devs += _get_ovs_bridge_port_devices(iface_state)
        elif iface_type == LB.TYPE:
            devs += bridge.get_slaves(nmdev)
        elif iface_type == bond.BOND_TYPE:
            devs += bond.get_slaves(nmdev)
    return devs


def _get_ovs_bridge_port_devices(iface_state):
    """
    Report a list of all ovs ports and interfaces that are connected to the
    OVS bridge.
    Note: Ports must be activated before the ifaces (NM limitation).
    """
    ifaces = [p['name'] for p in iface_state.get('bridge', {}).get('port', [])]
    ports = [ovs.PORT_PROFILE_PREFIX + iface for iface in ifaces]
    devnames = ports + ifaces
    nmdevs = []
    for devname in devnames:
        dev = device.get_device_by_name(devname)
        if dev:
            nmdevs.append(dev)
    return nmdevs


def prepare_proxy_ifaces_desired_state(ifaces_desired_state):
    """
    Prepare the state of the "proxy" interfaces. These are interfaces that
    exist as NM entities/profiles, but are invisible to the API.
    These proxy interfaces state is created as a side effect of other ifaces
    definition.
    Note: This function modifies the ifaces_desired_state content in addition
    to returning a new set of states for the proxy interfaces.

    In OVS case, the port profile is the proxy, it is not part of the public
    state of the system, but internal to the NM provider.
    """
    new_ifaces_desired_state = []
    for iface_desired_state in ifaces_desired_state:
        master_type = iface_desired_state.get(MASTER_TYPE_METADATA)
        if master_type != ovs.BRIDGE_TYPE:
            continue
        port_opts_metadata = iface_desired_state.get(BRPORT_OPTIONS_METADATA)
        if port_opts_metadata is None:
            continue
        port_options = ovs.translate_port_options(port_opts_metadata)
        port_iface_desired_state = _create_ovs_port_iface_desired_state(
            iface_desired_state, port_options
        )
        new_ifaces_desired_state.append(port_iface_desired_state)
        # The "visible" slave/interface needs to point to the port profile
        iface_desired_state[MASTER_METADATA] = port_iface_desired_state['name']
        iface_desired_state[MASTER_TYPE_METADATA] = ovs.PORT_TYPE
    return new_ifaces_desired_state


def _create_ovs_port_iface_desired_state(iface_desired_state, port_options):
    return {
        'name': ovs.PORT_PROFILE_PREFIX + iface_desired_state['name'],
        'type': ovs.PORT_TYPE,
        'state': iface_desired_state['state'],
        MASTER_METADATA: iface_desired_state[MASTER_METADATA],
        MASTER_TYPE_METADATA: iface_desired_state[MASTER_TYPE_METADATA],
        'options': port_options,
    }


def _build_connection_profile(iface_desired_state, base_con_profile=None):
    iface_type = translator.Api2Nm.get_iface_type(iface_desired_state['type'])

    base_profile = base_con_profile.profile if base_con_profile else None

    settings = [
        ipv4.create_setting(iface_desired_state.get('ipv4'), base_profile),
        ipv6.create_setting(iface_desired_state.get('ipv6'), base_profile),
    ]

    con_setting = connection.ConnectionSetting()
    if base_profile:
        con_setting.import_by_profile(base_con_profile)
    else:
        con_setting.create(
            con_name=iface_desired_state['name'],
            iface_name=iface_desired_state['name'],
            iface_type=iface_type,
        )
    master = iface_desired_state.get(MASTER_METADATA)
    _translate_master_type(iface_desired_state)
    master_type = iface_desired_state.get(MASTER_TYPE_METADATA)
    con_setting.set_master(master, master_type)
    settings.append(con_setting.setting)

    wired_setting = wired.create_setting(iface_desired_state, base_profile)
    if wired_setting:
        settings.append(wired_setting)

    user_setting = user.create_setting(iface_desired_state, base_profile)
    if user_setting:
        settings.append(user_setting)

    bond_opts = translator.Api2Nm.get_bond_options(iface_desired_state)
    if bond_opts:
        settings.append(bond.create_setting(bond_opts))
    elif iface_type == bridge.BRIDGE_TYPE:
        bridge_options = iface_desired_state.get('bridge', {}).get('options')
        if bridge_options:
            linux_bridge_setting = bridge.create_setting(
                bridge_options, base_profile
            )
            settings.append(linux_bridge_setting)
    elif iface_type == ovs.BRIDGE_TYPE:
        ovs_bridge_options = ovs.translate_bridge_options(iface_desired_state)
        if ovs_bridge_options:
            settings.append(ovs.create_bridge_setting(ovs_bridge_options))
    elif iface_type == ovs.PORT_TYPE:
        ovs_port_options = iface_desired_state.get('options')
        settings.append(ovs.create_port_setting(ovs_port_options))
    elif iface_type == ovs.INTERNAL_INTERFACE_TYPE:
        settings.append(ovs.create_interface_setting())

    bridge_port_options = iface_desired_state.get(BRPORT_OPTIONS_METADATA)
    if bridge_port_options and master_type == bridge.BRIDGE_TYPE:
        settings.append(
            bridge.create_port_setting(bridge_port_options, base_profile)
        )

    vlan_setting = vlan.create_setting(iface_desired_state, base_profile)
    if vlan_setting:
        settings.append(vlan_setting)

    new_profile = connection.ConnectionProfile()
    new_profile.create(settings)
    return new_profile


def _translate_master_type(iface_desired_state):
    """
    Translates the master type metadata names to their equivalent
    NM type names.
    """
    master_type = iface_desired_state.get(MASTER_TYPE_METADATA)
    if master_type == LB.TYPE:
        iface_desired_state[MASTER_TYPE_METADATA] = bridge.BRIDGE_TYPE



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import copy
import logging
import six

from libnmstate import metadata
from libnmstate import netinfo
from libnmstate import nm
from libnmstate import schema
from libnmstate import state
from libnmstate import sysctl
from libnmstate import validator
from libnmstate.error import NmstateConflictError
from libnmstate.error import NmstateLibnmError
from libnmstate.error import NmstatePermissionError
from libnmstate.error import NmstateValueError
from libnmstate.nm import nmclient


def apply(desired_state, verify_change=True, commit=True, rollback_timeout=60):
    """
    Apply the desired state

    :param verify_change: Check if the outcome state matches the desired state
        and rollback if not.
    :param commit: Commit the changes after verification if the state matches.
    :param rollback_timeout: Revert the changes if they are not commited within
        this timeout (specified in seconds).
    :type verify_change: bool
    :type commit: bool
    :type rollback_timeout: int (seconds)
    :returns: Checkpoint identifier
    :rtype: str
    """
    desired_state = copy.deepcopy(desired_state)
    validator.validate(desired_state)
    validator.validate_capabilities(desired_state, netinfo.capabilities())
    validator.validate_dhcp(desired_state)
    validator.validate_dns(desired_state)

    checkpoint = _apply_ifaces_state(
        state.State(desired_state), verify_change, commit, rollback_timeout
    )
    if checkpoint:
        return str(checkpoint.dbuspath)


def commit(checkpoint=None):
    """
    Commit a checkpoint that was received from `apply()`.

    :param checkpoint: Checkpoint to commit. If not specified, a checkpoint
        will be selected and committed.
    :type checkpoint: str
    """

    nmcheckpoint = _choose_checkpoint(checkpoint)
    try:
        nmcheckpoint.destroy()
    except nm.checkpoint.NMCheckPointError as e:
        raise NmstateValueError(str(e))


def rollback(checkpoint=None):
    """
    Roll back a checkpoint that was received from `apply()`.

    :param checkpoint: Checkpoint to roll back. If not specified, a checkpoint
        will be selected and rolled back.
    :type checkpoint: str
    """

    nmcheckpoint = _choose_checkpoint(checkpoint)
    try:
        nmcheckpoint.rollback()
    except nm.checkpoint.NMCheckPointError as e:
        raise NmstateValueError(str(e))


def _choose_checkpoint(dbuspath):
    if not dbuspath:
        candidates = nm.checkpoint.get_checkpoints()
        if candidates:
            dbuspath = candidates[0]

    if not dbuspath:
        raise NmstateValueError("No checkpoint specified or found")
    checkpoint = nm.checkpoint.CheckPoint(dbuspath=dbuspath)
    return checkpoint


def _apply_ifaces_state(
    desired_state, verify_change, commit, rollback_timeout
):
    current_state = state.State(netinfo.show())

    desired_state.sanitize_ethernet(current_state)
    desired_state.sanitize_dynamic_ip()
    desired_state.merge_routes(current_state)
    desired_state.merge_dns(current_state)
    metadata.generate_ifaces_metadata(desired_state, current_state)

    validator.validate_interfaces_state(desired_state, current_state)
    validator.validate_routes(desired_state, current_state)

    new_interfaces = _list_new_interfaces(desired_state, current_state)

    try:
        with nm.checkpoint.CheckPoint(
            autodestroy=commit, timeout=rollback_timeout
        ) as checkpoint:
            with _setup_providers():
                ifaces2add, ifaces_add_configs = _add_interfaces(
                    new_interfaces, desired_state
                )
                state2edit = _create_editable_desired_state(
                    desired_state, current_state, new_interfaces
                )
                ifaces2edit, ifaces_edit_configs = _edit_interfaces(state2edit)
                nm.applier.set_ifaces_admin_state(
                    ifaces2add + ifaces2edit,
                    con_profiles=ifaces_add_configs + ifaces_edit_configs,
                )
            _disable_ipv6(desired_state)
            if verify_change:
                _verify_change(desired_state)
        if not commit:
            return checkpoint
    except nm.checkpoint.NMCheckPointPermissionError:
        raise NmstatePermissionError('Error creating a check point')
    except nm.checkpoint.NMCheckPointCreationError:
        raise NmstateConflictError('Error creating a check point')


def _create_editable_desired_state(
    desired_state, current_state, new_intefaces
):
    """
    Create a new state object that includes only existing interfaces which need
    to be edited/changed.
    """
    state2edit = state.create_state(
        desired_state.state,
        interfaces_to_filter=(
            set(current_state.interfaces) - set(new_intefaces)
        ),
    )
    state2edit.merge_interfaces(current_state)
    return state2edit


def _list_new_interfaces(desired_state, current_state):
    return [
        name
        for name in six.viewkeys(desired_state.interfaces)
        - six.viewkeys(current_state.interfaces)
        if desired_state.interfaces[name].get(schema.Interface.STATE)
        not in (schema.InterfaceState.ABSENT, schema.InterfaceState.DOWN)
    ]


def _verify_change(desired_state):
    current_state = state.State(netinfo.show())
    desired_state.verify_interfaces(current_state)
    desired_state.verify_routes(current_state)
    desired_state.verify_dns(current_state)


@contextmanager
def _setup_providers():
    mainloop = nmclient.mainloop()
    yield
    success = mainloop.run(timeout=20)
    if not success:
        nmclient.mainloop(refresh=True)
        raise NmstateLibnmError(
            'Unexpected failure of libnm when running the mainloop: {}'.format(
                mainloop.error
            )
        )


def _add_interfaces(new_interfaces, desired_state):
    logging.debug('Adding new interfaces: %s', new_interfaces)

    ifaces2add = [desired_state.interfaces[name] for name in new_interfaces]

    ifaces2add += nm.applier.prepare_proxy_ifaces_desired_state(ifaces2add)
    ifaces_configs = nm.applier.prepare_new_ifaces_configuration(ifaces2add)
    nm.applier.create_new_ifaces(ifaces_configs)

    return (ifaces2add, ifaces_configs)


def _edit_interfaces(state2edit):
    logging.debug('Editing interfaces: %s', list(state2edit.interfaces))

    ifaces2edit = list(six.viewvalues(state2edit.interfaces))

    iface2prepare = list(
        filter(
            lambda state: state.get('state') not in ('absent', 'down'),
            ifaces2edit,
        )
    )
    proxy_ifaces = nm.applier.prepare_proxy_ifaces_desired_state(iface2prepare)
    ifaces_configs = nm.applier.prepare_edited_ifaces_configuration(
        iface2prepare + proxy_ifaces
    )
    nm.applier.edit_existing_ifaces(ifaces_configs)

    return (ifaces2edit + proxy_ifaces, ifaces_configs)


def _index_by_name(ifaces_state):
    return {iface['name']: iface for iface in ifaces_state}


def _disable_ipv6(desired_state):
    """
    Identify in the desired state all interfaces that explicitly disable
    the IPv6 stack and apply it through sysfs.

    This is an intermediate workaround for https://bugzilla.redhat.com/1643841.
    """
    for ifstate in six.viewvalues(desired_state.interfaces):
        ipv6_state = ifstate.get(schema.Interface.IPV6, {})
        if ipv6_state.get('enabled') is False:
            sysctl.disable_ipv6(ifstate[schema.Interface.NAME])



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import six


def get_slaves_from_state(state, default=()):
    ports = state.get('bridge', {}).get('port')
    if ports is None:
        return default
    return [p['name'] for p in ports]


def set_bridge_ports_metadata(master_state, slave_state):
    _set_common_slaves_metadata(master_state, slave_state)

    ports = master_state.get('bridge', {}).get('port', [])
    port = next(
        six.moves.filter(lambda n: n['name'] == slave_state['name'], ports), {}
    )
    slave_state['_brport_options'] = port


def _set_common_slaves_metadata(master_state, slave_state):
    slave_state['_master'] = master_state['name']
    slave_state['_master_type'] = master_state['type']



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from . import applier
from . import bond
from . import bridge
from . import checkpoint
from . import connection
from . import device
from . import dns
from . import ipv4
from . import ipv6
from . import nmclient
from . import ovs
from . import translator
from . import user
from . import vlan
from . import wired

applier
bond
bridge
checkpoint
connection
device
dns
ipv4
ipv6
nmclient
ovs
translator
user
vlan
wired



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import errno

from . import error


class ChangeIpv6StateError(error.NmstateError):
    pass


def enable_ipv6(dev):
    _change_ipv6_state(dev, disable=False)


def disable_ipv6(dev):
    _change_ipv6_state(dev, disable=True)


def _change_ipv6_state(dev, disable):
    try:
        with open('/proc/sys/net/ipv6/conf/%s/disable_ipv6' % dev, 'w') as f:
            f.write('1' if disable else '0')
    except IOError as e:
        if e.errno == errno.ENOENT and disable:
            # IPv6 stack is (already) not available on this device
            return
        raise ChangeIpv6StateError(str(e))


def is_disabled_ipv6(dev='default'):
    try:
        with open('/proc/sys/net/ipv6/conf/%s/disable_ipv6' % dev) as f:
            return int(f.read())
    except IOError as e:
        if e.errno == errno.ENOENT:
            return 1
        else:
            raise error.NmstateError(str(e))



import logging
import setuptools
from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()


def _is_outdated_setuptools():
    """
    Return True if setuptool is older than 20.5 which introduced environment
    markers for requirements:
        https://www.python.org/dev/peps/pep-0508/#environment-markers
    """
    major, minor = setuptools.__version__.split('.')[:2]
    return (int(major), int(minor)) < (20, 5)


def requirements():
    req = []
    setuptools_is_outdated = _is_outdated_setuptools()
    with open('requirements.txt') as fd:
        for line in fd:
            line.strip()
            if setuptools_is_outdated:
                if ';' in line:
                    # Remove the environment marker
                    logging.warning(
                        "Your setuptools is too old(<20.5) to support "
                        "environment marker, removing the environment marker "
                        "as workaround but it could be buggy"
                    )
                    line = line.split(';')[0]
            if not line.startswith('#'):
                req.append(line)
    return req


def get_version():
    with open('VERSION') as f:
        version = f.read().strip()
    return version


setup(
    name='nmstate',
    version=get_version(),
    description='Declarative network manager API',
    author="Edward Haas",
    author_email="ehaas@redhat.com",
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://nmstate.github.io/',
    license='GPLv2+',
    packages=find_packages(),
    install_requires=requirements(),
    entry_points={
        'console_scripts': ['nmstatectl = nmstatectl.nmstatectl:main']
    },
    package_data={'libnmstate': ['schemas/operational-state.yaml']},
)



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging

from . import nmclient


class AlternativeACState(object):
    UNKNOWN = 0
    ACTIVE = 1
    ACTIVATING = 2
    FAIL = 3


class ActivationError(Exception):
    pass


class ActiveConnection(object):
    def __init__(self, active_connection=None):
        self.handlers = set()
        self._act_con = active_connection
        self._mainloop = nmclient.mainloop()
        self._state = None

        nmdevs = None
        if active_connection:
            nmdevs = active_connection.get_devices()
        self._nmdev = nmdevs[0] if nmdevs else None

        if self._act_con:
            self.refresh_state()

    def import_by_device(self, nmdev=None):
        assert self._act_con is None

        if nmdev:
            self._nmdev = nmdev
        if self._nmdev:
            self._act_con = self._nmdev.get_active_connection()

    def deactivate(self):
        """
        Deactivating the current active connection,
        The profile itself is not removed.

        For software devices, deactivation removes the devices from the kernel.
        """
        self._mainloop.push_action(self._safe_deactivate_async)

    def _safe_deactivate_async(self):
        act_connection = self._nmdev.get_active_connection()
        mainloop = nmclient.mainloop()
        if not act_connection or act_connection.props.state in (
            nmclient.NM.ActiveConnectionState.DEACTIVATING,
            nmclient.NM.ActiveConnectionState.DEACTIVATED,
        ):
            # Nothing left to do here, call the next action.
            mainloop.execute_next_action()
            return

        user_data = None
        client = nmclient.client()
        client.deactivate_connection_async(
            act_connection,
            mainloop.cancellable,
            self._deactivate_connection_callback,
            user_data,
        )

    def _deactivate_connection_callback(self, src_object, result, user_data):
        try:
            success = src_object.deactivate_connection_finish(result)
        except Exception as e:
            if self._mainloop.is_action_canceled(e):
                logging.debug(
                    'Connection deactivation aborted on %s: error=%s',
                    self._nmdev.get_iface(),
                    e,
                )
            else:
                self._mainloop.quit(
                    'Connection deactivation failed on {}: error={}'.format(
                        self._nmdev.get_iface(), e
                    )
                )
            return

        if success:
            logging.debug(
                'Connection deactivation succeeded on %s',
                self._nmdev.get_iface(),
            )
            self._mainloop.execute_next_action()
        else:
            self._mainloop.quit(
                'Connection deactivation failed on %s: error=unknown'
                % self._nmdev.get_iface()
            )

    def refresh_state(self):
        self._state = self._act_con.get_state()
        self._state_reason = self._act_con.get_state_reason()
        self._alternative_state = AlternativeACState.UNKNOWN

        nm_acs = nmclient.NM.ActiveConnectionState
        nm_acsreason = nmclient.NM.ActiveConnectionStateReason
        if self._state == nm_acs.DEACTIVATED:
            unable_to_activate = (
                not self._nmdev
                or (
                    self._state_reason is not None
                    and self._state_reason != nm_acsreason.DEVICE_DISCONNECTED
                )
                or self._nmdev.get_active_connection() is not self._act_con
            )
            if unable_to_activate:
                self._alternative_state = AlternativeACState.FAIL
            # Use the device-state as an alternative to determine if active.
            elif (
                self.nmdev_state <= nmclient.NM.DeviceState.DISCONNECTED
                or self.nmdev_state > nmclient.NM.DeviceState.DEACTIVATING
            ):
                self._alternative_state = AlternativeACState.FAIL

    @property
    def is_active(self):
        nm_acs = nmclient.NM.ActiveConnectionState
        if self._state == nm_acs.ACTIVATED:
            return True
        elif self._state == nm_acs.ACTIVATING:
            # master connections qualify as activated once they
            # reach IP-Config state. That is because they may
            # wait for slave devices to attach
            return (
                _is_device_master_type(self._nmdev)
                and nmclient.NM.DeviceState.IP_CONFIG
                <= self.nmdev_state
                <= nmclient.NM.DeviceState.ACTIVATED
            )

        return False

    @property
    def is_activating(self):
        activation_failed = self._alternative_state == AlternativeACState.FAIL
        return not self.is_active and not activation_failed

    @property
    def reason(self):
        return self._state_reason

    @property
    def nm_active_connection(self):
        return self._act_con

    @property
    def devname(self):
        return self._nmdev.get_iface()

    @property
    def nmdevice(self):
        return self._nmdev

    @nmdevice.setter
    def nmdevice(self, nmdev):
        assert self._nmdev is None
        self._nmdev = nmdev

    @property
    def state(self):
        return self._state

    @property
    def nmdev_state(self):
        return (
            self._nmdev.get_state()
            if self._nmdev
            else nmclient.NM.DeviceState.UNKNOWN
        )


def _is_device_master_type(nmdev):
    if nmdev:
        gobject = nmclient.GObject
        is_master_type = (
            gobject.type_is_a(nmdev, nmclient.NM.DeviceBond)
            or gobject.type_is_a(nmdev, nmclient.NM.DeviceBridge)
            or gobject.type_is_a(nmdev, nmclient.NM.DeviceTeam)
            or gobject.type_is_a(nmdev, nmclient.NM.DeviceOvsBridge)
        )
        return is_master_type
    return False



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging

import dbus

import libnmstate.nm.nmclient


DBUS_STD_PROPERTIES_IFNAME = 'org.freedesktop.DBus.Properties'

CHECKPOINT_CREATE_FLAG_DESTROY_ALL = 0x01
CHECKPOINT_CREATE_FLAG_DELETE_NEW_CONNECTIONS = 0x02
CHECKPOINT_CREATE_FLAG_DISCONNECT_NEW_DEVICES = 0x04

NM_PERMISSION_DENIED = 'org.freedesktop.NetworkManager.PermissionDenied'


_nmdbus_manager = None


class NMCheckPointError(Exception):
    pass


class NMCheckPointCreationError(NMCheckPointError):
    pass


class NMCheckPointPermissionError(NMCheckPointError):
    pass


def nmdbus_manager():
    """
    Returns the NM manager.
    If it does not exists, it will initialize the dbus connection and
    create the manager.
    """
    global _nmdbus_manager
    if _nmdbus_manager is None:
        _NMDbus.init()
        _nmdbus_manager = _NMDbusManager()
    return _nmdbus_manager


class _NMDbus(object):
    BUS_NAME = 'org.freedesktop.NetworkManager'

    bus = None

    @staticmethod
    def init():
        _NMDbus.bus = dbus.SystemBus()


class _NMDbusManager(object):
    IF_NAME = 'org.freedesktop.NetworkManager'
    OBJ_PATH = '/org/freedesktop/NetworkManager'

    def __init__(self):
        mng_proxy = _NMDbus.bus.get_object(
            _NMDbus.BUS_NAME, _NMDbusManager.OBJ_PATH
        )
        self.properties = dbus.Interface(mng_proxy, DBUS_STD_PROPERTIES_IFNAME)
        self.interface = dbus.Interface(mng_proxy, _NMDbusManager.IF_NAME)


def get_checkpoints():
    nmclient = libnmstate.nm.nmclient.client(refresh=True)

    return [c.get_path() for c in nmclient.get_checkpoints()]


class CheckPoint(object):
    def __init__(self, timeout=60, autodestroy=True, dbuspath=None):
        self._manager = nmdbus_manager()
        self._timeout = timeout
        self._dbuspath = dbuspath
        self._autodestroy = autodestroy

    def __enter__(self):
        self.create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            if self._autodestroy:
                # If the rollback has been explicit or implicit triggered
                # already, this command should fail.
                self.destroy()
        else:
            self.rollback()

    def create(self):
        devs = []
        timeout = self._timeout
        cp_flags = (
            CHECKPOINT_CREATE_FLAG_DELETE_NEW_CONNECTIONS
            | CHECKPOINT_CREATE_FLAG_DISCONNECT_NEW_DEVICES
        )
        try:
            dbuspath = self._manager.interface.CheckpointCreate(
                devs, timeout, cp_flags
            )
            logging.debug(
                'Checkpoint %s created for all devices: %s', dbuspath, timeout
            )
            self._dbuspath = dbuspath
        except dbus.exceptions.DBusException as e:
            if e.get_dbus_name() == NM_PERMISSION_DENIED:
                raise NMCheckPointPermissionError(str(e))
            raise NMCheckPointCreationError(str(e))

    def destroy(self):
        try:
            self._manager.interface.CheckpointDestroy(self._dbuspath)
        except dbus.exceptions.DBusException as e:
            raise NMCheckPointError(str(e))

        logging.debug('Checkpoint %s destroyed', self._dbuspath)

    def rollback(self):
        try:
            result = self._manager.interface.CheckpointRollback(self._dbuspath)
        except dbus.exceptions.DBusException as e:
            raise NMCheckPointError(str(e))
        logging.debug(
            'Checkpoint %s rollback executed: %s', self._dbuspath, result
        )
        return result

    @property
    def dbuspath(self):
        return self._dbuspath
